"""empty message

Revision ID: roc_dingo_0029_tm_log_partitioning
Revises: roc_dingo_0028_modify_sbm_log_status_retrieved
Create Date: 2023-11-02 19:14:28.193974

"""

from pathlib import Path

from alembic import op

from poppy.pop.alembic.helpers import execute, user

# revision identifiers, used by Alembic.
revision = "roc_dingo_0029_tm_log_partitioning"
down_revision = "roc_dingo_0028_modify_sbm_log_status_retrieved"
branch_labels = None
depends_on = None


def upgrade():
    # This command works
    filename = "SQL/roc_dingo_0029__upgrade.sql"
    sql_file = Path(__file__).parent / filename
    with open(sql_file, "r") as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)

    # added by hand
    # fix table permissions
    grant_user_perm = """GRANT SELECT, INSERT, UPDATE
        ON ALL TABLES IN SCHEMA pipeline TO {0}""".format(user)
    execute(grant_user_perm)
    # fix truncate table permissions
    # (required when running event_to_db task with --force option)
    grant_user_truncate = """GRANT TRUNCATE
        ON TABLE pipeline.event_log TO {0}""".format(user)
    execute(grant_user_truncate)
    # Make sure user can reset event_log_id_seq
    # alter_seq_user = (
    #    """ALTER SEQUENCE IF EXISTS event_log_id_seq
    #    OWNER TO {0}""".format(user)
    # )
    # execute(alter_seq_user)
    # fix sequence permissions
    grant_user_seq = """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    execute(grant_user_seq)


def downgrade():
    # This command works
    filename = "SQL/roc_dingo_0029__downgrade.sql"
    sql_file = Path(__file__).parent / filename
    with open(sql_file, "r") as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)

    # added by hand
    # fix table permissions
    grant_user_perm = """GRANT SELECT, INSERT, UPDATE
        ON ALL TABLES IN SCHEMA pipeline TO {0}""".format(user)
    execute(grant_user_perm)
    # Make sure admin can reset event_log_id_seq
    # alter_seq_admin = (
    #    """ALTER SEQUENCE IF EXISTS event_log_id_seq
    #    OWNER TO {0}""".format(admin)
    # )
    # execute(alter_seq_admin)
    # Fix sequence permissions
    grant_user_seq = """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    execute(grant_user_seq)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""

import pytest
import os.path as osp

from poppy.core.test import CommandTestCase

from roc.dingo.models.file import FileLog
from roc.dingo.tests.test_log_file import DingoLogFileTest


# Tests on roc.dingo.tasks.log_file methods
class TestDingoLogFileToRocdb(CommandTestCase):
    def setup_method(self, method):
        super().setup_method(method)

        # empty the database
        # DingoTest.create_db(self)
        self.session.query(FileLog).delete()
        self.session.flush()

    def teardown_method(self, method):
        super().teardown_method(method)

        # close and delete database
        self.session.close()
        # DingoTest.clear_db(self)

    @pytest.mark.parametrize(
        "root, date, level, nb_files",
        [
            ("logfile_to_rocdb/DATA-1", None, None, 61),
            ("logfile_to_rocdb/DATA-1", "2020/08/19", None, 31),
            ("logfile_to_rocdb/DATA-1", "2020/08/19", ["L0"], 1),
            ("logfile_to_rocdb/DATA-1", "2020/08/19", ["L0", "L1"], 14),
            ("logfile_to_rocdb/DATA-1", "2020/08/19", ["L2"], 0),
            ("logfile_to_rocdb/DATA-1", "2020/08", ["L2"], 17),
        ],
    )
    def test_logfile_to_rocdb(self, root, date, level, nb_files):
        """
        Test insertion with different input parameters
        Check the number of files inserted in db after each
        """
        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)

        # --- initialize the command ---
        main_command = ["pop", "-ll", "INFO", "dingo", "logfile_to_rocdb"]

        if root is not None:
            main_command += ["--root", root]

        if date is not None:
            main_command += ["--date", date]

        if level is not None:
            main_command += ["--level"] + level

        # --- run the command ---
        self.run_command(main_command)

        # --- make assertions ---

        results = self.session.query(FileLog).all()
        assert len(results) == nb_files

        # self.session.close()

    @pytest.mark.parametrize(
        "root",
        [
            ("logfile_to_rocdb/DATA-1"),
        ],
    )
    def test_logfile_to_rocdb_multiple(self, root):
        """
        Test multiple insertion with different input parameters
        Check the number of files inserted in db after each
        """

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)

        # --- initialize the commands ---
        commands = [
            [
                "pop",
                "-ll",
                "INFO",
                "dingo",
                "logfile_to_rocdb",
                "--root",
                root,
                "--date",
                "2020/08/19",
                "--level",
                "L0",
                "L1",
            ],
            [
                "pop",
                "-ll",
                "INFO",
                "dingo",
                "logfile_to_rocdb",
                "--root",
                root,
                "--date",
                "2020/08/19",
                "--level",
                "L0",
            ],
            [
                "pop",
                "-ll",
                "INFO",
                "dingo",
                "logfile_to_rocdb",
                "--root",
                root,
                "--date",
                "2020/08",
                "--level",
                "L2",
            ],
        ]

        nb_files = [14, 14, 31]

        for i, cmd in enumerate(commands):
            # --- run the  command ---
            self.run_command(cmd)

            # --- make assertions ---
            results = self.session.query(FileLog).all()
            assert len(results) == nb_files[i]

        # self.session.close()

    @pytest.mark.parametrize(
        "root",
        [
            ("logfile_to_rocdb/SK-1"),
        ],
    )
    def test_logsk_to_rocdb(self, root):
        """
        Test spice kernel insertion
        Check the number of files inserted in db after each
        """
        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)

        # --- initialize the commands ---
        command = ["pop", "-ll", "INFO", "dingo", "logsk_to_rocdb", "--root", root]

        nb_files = 5

        # --- run the  command ---
        self.run_command(command)

        # --- make assertions ---
        results = self.session.query(FileLog).filter(FileLog.level == "SK").all()
        assert len(results) == nb_files

        # self.session.close()

    @pytest.mark.parametrize(
        "filesroot, skroot, expected_before, expected_after",
        [("logfile_to_rocdb/DATA-1", "logfile_to_rocdb/SK-1", 32, 12)],
    )
    def test_update_missing_parents_in_db(
        self, filesroot, skroot, expected_before, expected_after
    ):
        """
        Test spice kernel insertion
        Check the number of files inserted in db after each
        """

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        filesroot = osp.join(data_test_path, filesroot)
        skroot = osp.join(data_test_path, skroot)

        command = [
            "pop",
            "-ll",
            "INFO",
            "dingo",
            "logfile_to_rocdb",
            "--root",
            filesroot,
        ]
        self.run_command(command)
        command = ["pop", "-ll", "INFO", "dingo", "logsk_to_rocdb", "--root", skroot]
        self.run_command(command)

        results = (
            self.session.query(FileLog)
            .filter(FileLog.error_log.like("%Missing parents%"))
            .all()
        )
        assert len(results) == expected_before

        command = [
            "pop",
            "dingo",
            "-ll",
            "INFO",
            "update_missing_parents_in_db",
            "--root",
            filesroot,
        ]
        self.run_command(command)

        results = (
            self.session.query(FileLog)
            .filter(FileLog.error_log.like("%Missing parents%"))
            .all()
        )
        assert len(results) == expected_after

        # self.session.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

### Mandatory code in every plugin.models __init__.py needed for alembic
### DO NOT EDIT

import importlib
import pkgutil
import os
import sys
from poppy.core.db.base import Base

pkg_dir = os.path.dirname(__file__)
for module_loader, name, ispkg in pkgutil.iter_modules([pkg_dir]):
    importlib.import_module("." + name, __package__)

# Creates a list of all the tables represented by classes in the current plugin
# It is needed for alembic to generate migrations correctly
tables = {
    cls.__tablename__: cls
    for cls in Base.__subclasses__()
    if sys.modules[__name__].__name__ in cls.__module__
}

### End of mandatory code



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""

import pytest
import os.path as osp
from random import randrange, sample
from argparse import Namespace
from unittest import mock

from poppy.core.pipeline import Pipeline
from poppy.core.logger import logger
from poppy.core.test import TaskTestCase

from roc.dingo.tasks import CheckSOARDataAvailability
from roc.dingo.models.file import FileLog
from roc.dingo.tests.test_log_file import DingoLogFileTest


# Tests on roc.dingo.tasks.log_file methods
class TestDingoLogFileCheckSOARDataAvailability(TaskTestCase):
    @mock.patch("roc.dingo.tasks.CheckSOARDataAvailability.soar_run_query")
    @pytest.mark.parametrize(
        "root, data_file, public_data_file, nb_to_archive",
        [
            (
                "logfile_to_rocdb/DATA-1/",
                "roc_data_files_all.dat",
                "roc_data_files_public_filename.dat",
                31,
            )
        ],
    )
    def test_store_archived(
        self, mock_query, root, data_file, public_data_file, nb_to_archive
    ):
        """
        Start from an empty database
        populates it by hand with files from data_file
        and public filename from public_data_file
        mock the respone from the TAP service

        :return:
        """

        logger.info("test_store_archived() ")

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(url="fakeurl", dry_run=False)

        # --- database setup ---
        session = DingoLogFileTest.setup_session()

        # --- define data test --
        data1 = DingoLogFileTest.open_test_data(data_file, root)

        # files with their public_filename
        public_filename = DingoLogFileTest.read_data(public_data_file)
        public_filename = [line.split() for line in public_filename]
        public_filename = {line[0]: line[1] for line in public_filename}
        delivered_files = []

        # populate database
        for item, data in data1.items():
            file_log = DingoLogFileTest.data_to_file_log(data, root)
            if file_log.basename in public_filename:
                file_log.public_filename = public_filename[file_log.basename]
                file_log.is_delivered = True
                delivered_files.append([file_log.public_filename])

            session.add(file_log)
        session.flush()

        # --- get database entries ---
        db_before = (
            session.query(FileLog).filter(FileLog.public_filename.isnot(None)).all()
        )

        # create the mocked response from the TAP service
        fake_nb_archived = randrange(len(delivered_files))
        fake_archived = {"data": sample(delivered_files, fake_nb_archived)}

        # --- define mock session result
        mock_query.return_value = fake_archived

        # --- initialise task & pipeline --
        self.task = CheckSOARDataAvailability()
        self.task.pipeline = Pipeline(args)
        self.task.instantiate_targets()
        self.task.session = session

        # --- initialise inputs ---

        # --- run the task ---
        self.run_task()

        # --- get database entries ---
        db_after = session.query(FileLog).filter(FileLog.is_archived == True).all()  # noqa : E712

        # --- get results ---
        store_results = self.task.outputs["store_archived"].data

        # --- make assertions ---

        assert len(db_before) == nb_to_archive
        assert len(db_after) == fake_nb_archived
        assert store_results["files_to_archive"] == nb_to_archive
        assert store_results["files_archived"] == fake_nb_archived

        # --- close session
        session.close()



# -*- coding: utf-8 -*-
"""
Add new column product that will receive name
without version and file extension

Revision ID: roc_dingo_0018_add_product_to_file_log
Revises: roc_dingo_0017_add_efecsevents
Create Date: 2021-11-29 16:11:25.270861

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "roc_dingo_0018b_add_product_to_file_log"
down_revision = "roc_dingo_0018a_task_delivered"
branch_labels = None
depends_on = None


def upgrade():
    # Add the new column product
    op.add_column(
        "file_log",
        sa.Column("product", sa.String(length=512), nullable=True),
        schema="pipeline",
    )

    # Fill it
    op.execute("""
        UPDATE pipeline.file_log
        SET product = regexp_replace(
            basename, '^([^\\.]+)_V[0-9U]+\\.[^$]+$', '\\1')
        """)

    # Set nullable=False after filling
    op.alter_column(
        "file_log",
        "product",
        existing_type=sa.VARCHAR(),
        nullable=False,
        schema="pipeline",
    )


def downgrade():
    op.drop_column("file_log", "product", schema="pipeline")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Functions used in all DINGO FileLog tests
"""

import os
import re
from datetime import datetime

from poppy.core.logger import logger
# from poppy.core.test import TaskTestCase, CommandTestCase

from roc.dingo.models.file import FileLog
from roc.dingo.constants import DATA_ALLOWED_EXTENSIONS, SPICE_KERNEL_ALLOWED_EXTENSIONS
from roc.dingo.tests.test_dingo import DingoTest


# class DingoLogFileTaskTestCase(TaskTestCase):
# @classmethod
# def setup_class(cls):
#    logger.debug("setup_class DingoLogFileTaskTestCase")
#    d = DingoLogFileTest()
#    d.get_test_data()

#     # --- database setup ---
#     session = DingoTestLogFile.setup_session()

# @classmethod
# def teardown_class(cls):
#     logger.debug('teardown_class DingoTaskTestCase')

#     # --- close session
#     session.close()


# class DingoLogFileCommandTestCase(CommandTestCase):
#    @classmethod
#    def setup_class(cls):
#        logger.debug("setup_class DingoLogFileCommandTestCase")
#        d = DingoLogFileTest()
#        d.get_test_data()


class DingoLogFileTest(DingoTest):
    #    base_url = f"{DingoTest.base_url}/logfile_to_rocdb"
    #    base_path = f"{DingoTest.base_path}/logfile_to_rocdb"

    @staticmethod
    def open_test_data(expected_filename, root):
        """
        Open <expected_filename> file, which contains a list of
            file1 size1 timestamp1
            file2 size2 timestamp2
            ...
        And creates a dictionary array

            :param expected_filename: filename containing expected data
            :param root: the data root in the file system
            :return: a dictionary array
        """

        # open the data_files file with data for test
        path = os.path.abspath(os.path.dirname(__file__))
        path = os.path.join(path, "data-tasks", expected_filename)
        expected_data = {}
        with open(path, "r") as f:
            # read a line corresponding to a file
            for line in f.readlines():
                relpath, size, cdate = line.split()
                basename = os.path.basename(relpath)
                # append the data
                expected_data[basename] = {
                    "filepath": os.path.join(root, relpath),
                    "size": int(size),
                    "creation_time": datetime.fromtimestamp(int(cdate), None),
                }

        return expected_data

    @staticmethod
    def data_to_file_log(data, root):
        dirname = os.path.dirname(data["filepath"]).replace(root, "")
        basename = os.path.basename(data["filepath"])
        file_path = dirname.split("/")
        file_name, file_extension = os.path.splitext(basename)

        is_data = True
        re_ext = "|".join(DATA_ALLOWED_EXTENSIONS)

        if file_extension.replace(".", "") in SPICE_KERNEL_ALLOWED_EXTENSIONS:
            is_data = False
            re_ext = "|".join(SPICE_KERNEL_ALLOWED_EXTENSIONS)

        if is_data:
            level = file_path[0]
            res = re.search(r"(v|V)([0-9a-zA-Z]+)\.(" + re_ext + ")", basename)
            res_id = 2
        else:
            level = "SK"
            res = re.search(r"([0-9]+)\.(" + re_ext + ")", basename)
            res_id = 1

        try:
            version = res.group(res_id)
        except Exception as e:
            logger.debug(e)
            version = "0"

        product = re.sub(r"^([^\.]+)_V[0-9U]+\.[^$]+$", r"\1", basename)

        file_log = {
            "basename": basename,
            "product": product,
            "dirname": dirname,
            "size": data["size"],
            "creation_time": data["creation_time"],
            "version": version,
            "state": "OK",
            "status": "Terminated",
            "level": level,
            "insert_time": datetime.now(),
            "is_archived": False,
            "is_latest": True,
        }
        return FileLog(**file_log)

    @staticmethod
    def setup_session():
        session = DingoTest.setup_session()

        # empty the table FileLog
        session.query(FileLog).delete()
        session.flush()

        return session



"""empty message

Revision ID: roc_dingo_0032_up_cuc_constraints
Revises: roc_dingo_0031_add_sbm_subtype
Create Date: 2025-02-06 10:51:57.794747

"""

from pathlib import Path
from alembic import op

# revision identifiers, used by Alembic.
revision = "roc_dingo_0032_up_cuc_constraints"
down_revision = "roc_dingo_0031_add_sbm_subtype"
branch_labels = None
depends_on = None


def upgrade():

    # Command to ensure there is no [cuc_time, sweep_step] or [cuc_time, kcoeff_pkt_nr] doublons before creating the constraint
    filename = "SQL/roc_dingo_0032__upgrade.sql"
    sql_file = Path(__file__).parent / filename
    with open(sql_file, "r") as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "bia_sweep_log_sweep_step_utc_time_key",
        "bia_sweep_log",
        schema="pipeline",
        type_="unique",
    )
    op.create_unique_constraint(
        "bia_sweep_log_cuc_time_sweep_step_key", "bia_sweep_log", ["cuc_time", "sweep_step"], schema="pipeline"
    )
    op.drop_constraint(
        "lfr_kcoeff_dump_utc_time_kcoeff_pkt_nr_key",
        "lfr_kcoeff_dump",
        schema="pipeline",
        type_="unique",
    )
    op.create_unique_constraint(
        "lfr_kcoeff_dump_cuc_time_kcoeff_pkt_nr_key", "lfr_kcoeff_dump", ["cuc_time", "kcoeff_pkt_nr"], schema="pipeline"
    )
    # ### end Alembic commands ###


def downgrade():

    # Command to ensure there is no [utc_time, sweep_step] or [utc_time, kcoeff_pkt_nr] doublons before creating the constraint
    filename = "SQL/roc_dingo_0032__downgrade.sql"
    sql_file = Path(__file__).parent / filename
    with open(sql_file, "r") as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("lfr_kcoeff_dump_cuc_time_kcoeff_pkt_nr_key", "lfr_kcoeff_dump", schema="pipeline", type_="unique")
    op.create_unique_constraint(
        "lfr_kcoeff_dump_utc_time_kcoeff_pkt_nr_key",
        "lfr_kcoeff_dump",
        ["utc_time", "kcoeff_pkt_nr"],
        schema="pipeline",
    )
    op.drop_constraint("bia_sweep_log_cuc_time_sweep_step_key", "bia_sweep_log", schema="pipeline", type_="unique")
    op.create_unique_constraint(
        "bia_sweep_log_sweep_step_utc_time_key",
        "bia_sweep_log",
        ["sweep_step", "utc_time"],
        schema="pipeline",
    )
    # ### end Alembic commands ###



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to insert EFECS event data into the ROC database."""

from datetime import datetime

import xmltodict

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import FileTarget

from roc.dingo.constants import (
    PIPELINE_DATABASE,
    TRYOUTS,
    TIME_WAIT_SEC,
    TIME_EFECS_STRFORMAT,
)
from roc.dingo.models.data import EfecsEvents
from roc.dingo.tools import get_columns, insert_in_db, query_db, glob_paths

__all__ = ["EfecsToDb"]


class EfecsToDb(Task):
    """
    Insert EFECS event data
    into the ROC database
    """

    plugin_name = "roc.dingo"
    name = "efecs_to_db"

    def add_targets(self):
        self.add_input(
            identifier="efecs_files",
            many=True,
            filepath=EfecsToDb.get_efecs_files,
            target_class=FileTarget,
        )

    @staticmethod
    def get_efecs_files(pipeline):
        try:
            efecs_files = pipeline.args.efecs_files
            if not isinstance(efecs_files, list):
                efecs_files = [efecs_files]
            return efecs_files
        except Exception as e:
            # If not defined as input argument, then assume that it is already
            # defined as target input
            logger.debug(e)
            pass

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):
        # get the input files
        self.efecs_files = glob_paths(self.inputs["efecs_files"].filepath)

        # Get tryouts from pipeline properties
        self.tryouts = self.pipeline.get("tryouts", default=[TRYOUTS], create=True)[0]

        # Get wait from pipeline properties
        self.wait = self.pipeline.get("wait", default=[TIME_WAIT_SEC], create=True)[0]

        # get a database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get table class
        self.model = EfecsEvents

        # Get table columns (remove primary key)
        self.columns = get_columns(self.model, remove=["id"])

    def run(self):
        logger.debug("Task EfecsToDb is starting")
        try:
            self.setup_inputs()
        except Exception as e:
            logger.error("Initializing inputs for task EfecsToDb has failed!")
            logger.debug(e)
            self.pipeline.exit()
            return

        # Number of EFECS files
        n_files = len(self.efecs_files)
        logger.info(f"{n_files} EFECS files to process")

        # loop over each input E-FECS file
        new_insert_count = 0
        already_insert_count = 0
        failed_insert_count = 0
        for i, current_file in enumerate(self.efecs_files):
            self.current_file = current_file
            logger.info(f"Processing {current_file}   ({n_files - i - 1} remaining)")
            # Parse file
            try:
                efecs_data = parse_efecs(current_file)

                # Make sure to always use list for EFECS events
                # And get total number of events in the EFECS file
                event_count = 0
                for key, val in efecs_data["events"].items():
                    if not isinstance(val, list):
                        efecs_data["events"][key] = [val]
                    event_count += len(efecs_data["events"][key])
            except Exception as e:
                logger.error(f"Parsing {current_file} has failed!")
                logger.debug(e)
                continue
            else:
                # GET LTP counter from input filename (e.g., "EFECS_M05.xml")
                ltp_count = int(efecs_data["header2"]["@SPKT_ltp_number"])

                # Get EFECS file generation time
                gen_time = datetime.strptime(
                    efecs_data["header"]["@gen_time"], TIME_EFECS_STRFORMAT
                )

            # Check database content against EFECS file info
            if self.is_uptodate(ltp_count, gen_time, event_count):
                logger.info(
                    f"{event_count} EFECS events from {current_file} already inserted, skip it"
                )
                continue
            else:
                logger.info(
                    f"{event_count} EFECS events will be inserted from {current_file}"
                )

            # Loops over each EFECS event found in the input file
            for key, val in efecs_data["events"].items():
                for current_entry in val:
                    data_to_insert = {"name": key}
                    data_to_insert["utc_time"] = datetime.strptime(
                        current_entry.pop("@time"), TIME_EFECS_STRFORMAT
                    )

                    data_to_insert["attributes"] = {
                        attr_key.replace("@", ""): attr_val
                        for attr_key, attr_val in current_entry.items()
                    }
                    data_to_insert["ltp_count"] = ltp_count
                    data_to_insert["gen_time"] = gen_time

                    # Define columns to check in case of update
                    update_fields_kwargs = {
                        col_name: col_val
                        for col_name, col_val in data_to_insert.items()
                        if col_name in ["name", "utc_time"]
                    }

                    # Insert / update efecs data into the database
                    status = insert_in_db(
                        self.session,
                        self.model,
                        data_to_insert,
                        update_fields=data_to_insert,
                        update_fields_kwargs=update_fields_kwargs,
                        tryouts=self.tryouts,
                        wait=self.wait,
                    )
                    if status < 0:
                        logger.error(
                            f"Cannot insert {data_to_insert} from {current_file}!"
                        )
                        failed_insert_count += 1
                    elif status == 0:
                        new_insert_count += 1
                        logger.debug(f"{data_to_insert} inserted")
                    else:
                        already_insert_count += 1
                        logger.debug(f"{data_to_insert} already inserted (updated)")

        if new_insert_count > 0:
            logger.info(f"{new_insert_count} EFECS events inserted")
        if already_insert_count > 0:
            logger.info(f"{already_insert_count} EFECS events updated")
        if failed_insert_count > 0:
            logger.warning(f"{failed_insert_count} EFECS events insertion have failed!")

    def is_uptodate(self, ltp_count, gen_time, event_count):
        """
        Compare content of the pipeline.efecs_events table with
        input EFECS file ltp count, generation time and number of events.
        Return True if the database and file content are the same.

        :param ltp_count: EFECS file LTP count (integer)
        :param gen_time: EFECS file generation time (datetime.datetime object)
        :param event_count: Number of EFECS events in the file
        :return: True if EFECS file events are already found in the database, False otherwise
        """
        is_found = False
        try:
            # Query database to get EFECS events with
            # input values for ltp_count and gen_time
            filters = EfecsEvents.ltp_count == ltp_count
            rows = query_db(
                self.session,
                EfecsEvents.gen_time,
                filters=filters,
                tryouts=self.tryouts,
                wait=self.wait,
                to_dict="records",
            )

            row_count = len(rows)
            if row_count == 0:
                logger.debug("No entry found for {ltp_count} in the database")
            else:
                is_found = (
                    all(current_event["gen_time"] == gen_time for current_event in rows)
                    and row_count == event_count
                )

            if not is_found and row_count > 0 and gen_time < rows[0]["gen_time"]:
                logger.warning(
                    f"Input EFECS file generation time ({gen_time}) for {self.current_file}"
                    f" is older than in database ({rows[0]['gen_time']})!"
                )
        except Exception as e:
            # If database cannot compare, assume is_found=False by default
            logger.error("Database content cannot be compared!")
            logger.debug(e)

        return is_found


def parse_efecs(efecs_file):
    """
    Parse EFECS XML input file.

    :param efecs_file: Path to EFECS file to parse
    :return: Content of input XML as a dictionary
    """
    with open(efecs_file, "r") as xml:
        xml_dict = xmltodict.parse(xml.read())

    return xml_dict["eventfile"]



# -*- coding: utf-8 -*-
"""Add a columns to store if file has already been delivered to ESA
and with which filename

Revision ID: roc_dingo_0016_add_is_delivered
Revises: roc_dingo_0015_add_dataqueue
Create Date: 2021-10-12 11:58:35.441300

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "roc_dingo_0018a_task_delivered"
down_revision = "roc_dingo_0017_add_efecsevents"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "file_log",
        sa.Column("is_delivered", sa.BOOLEAN(), nullable=True),
        schema="pipeline",
    )

    op.add_column(
        "file_log",
        sa.Column("public_filename", sa.String(), nullable=True),
        schema="pipeline",
    )

    op.add_column(
        "file_log",
        sa.Column("public_dirname", sa.String(), nullable=True),
        schema="pipeline",
    )

    # adding a column with a default value is not possible with Alembic
    # so create the column without the nullable constraint
    # fill the table with the default value (false)
    # after, add the nullable constraint

    op.execute("UPDATE pipeline.file_log SET is_delivered = false")
    op.alter_column("file_log", "is_delivered", nullable=False, schema="pipeline")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_column("file_log", "is_delivered", schema="pipeline")
    op.drop_column("file_log", "public_filename", schema="pipeline")
    op.drop_column("file_log", "public_dirname", schema="pipeline")
    # ### end Alembic commands ###



"""empty message

Revision ID: roc_dingo_0016_add_kcoeff_biascurr
Revises: roc_dingo_0015_add_dataqueue
Create Date: 2021-10-10 12:40:44.668707

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import (
    create_table,
    execute,
    user,
    drop_table,
)

# revision identifiers, used by Alembic.
revision = "roc_dingo_0016_add_kcoeffbiacur"
down_revision = "roc_dingo_0015_add_processqueue"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    create_table(
        "bia_current_log",
        sa.Column("id", sa.BIGINT(), nullable=False),
        sa.Column("tc_name", sa.String(length=256), nullable=False),
        sa.Column("utc_time", postgresql.TIMESTAMP(), nullable=False),
        sa.Column("bia_current", sa.FLOAT(), nullable=False),
        sa.Column("insert_time", postgresql.TIMESTAMP(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("utc_time", "tc_name"),
        schema="pipeline",
    )
    create_table(
        "lfr_kcoeff_dump",
        sa.Column("id", sa.BIGINT(), nullable=False),
        sa.Column("utc_time", postgresql.TIMESTAMP(), nullable=False),
        sa.Column("cuc_time", sa.String(length=1024), nullable=False),
        sa.Column("utc_time_is_predictive", sa.BOOLEAN(), nullable=True),
        sa.Column("kcoeff_pkt_cnt", sa.SMALLINT(), nullable=False),
        sa.Column("kcoeff_pkt_nr", sa.SMALLINT(), nullable=False),
        sa.Column("kcoeff_blk_nr", sa.SMALLINT(), nullable=False),
        sa.Column(
            "kcoeff_values", postgresql.JSONB(astext_type=sa.Text()), nullable=False
        ),
        sa.Column("insert_time", postgresql.TIMESTAMP(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("utc_time", "kcoeff_pkt_nr"),
        schema="pipeline",
    )
    op.alter_column(
        "solo_hk_param",
        "description",
        existing_type=sa.VARCHAR(),
        nullable=False,
        schema="pipeline",
    )
    op.alter_column(
        "solo_hk_param",
        "eng_value",
        existing_type=sa.VARCHAR(),
        nullable=False,
        schema="pipeline",
    )
    op.alter_column(
        "solo_hk_param",
        "raw_value",
        existing_type=sa.VARCHAR(),
        nullable=False,
        schema="pipeline",
    )
    op.alter_column(
        "solo_hk_param",
        "unit",
        existing_type=sa.VARCHAR(),
        nullable=False,
        schema="pipeline",
    )
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    execute(grant_user_seq)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "solo_hk_param",
        "unit",
        existing_type=sa.VARCHAR(),
        nullable=True,
        schema="pipeline",
    )
    op.alter_column(
        "solo_hk_param",
        "raw_value",
        existing_type=sa.VARCHAR(),
        nullable=True,
        schema="pipeline",
    )
    op.alter_column(
        "solo_hk_param",
        "eng_value",
        existing_type=sa.VARCHAR(),
        nullable=True,
        schema="pipeline",
    )
    op.alter_column(
        "solo_hk_param",
        "description",
        existing_type=sa.VARCHAR(),
        nullable=True,
        schema="pipeline",
    )
    drop_table("lfr_kcoeff_dump", schema="pipeline")
    drop_table("bia_current_log", schema="pipeline")
    # ### end Alembic commands ###



"""empty message

Revision ID: roc_dingo_0014_add_solohk
Revises: roc_dingo_0013_fix_db
Create Date: 2021-09-14 16:41:50.725120

"""

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import (
    create_table,
    execute,
    user,
    drop_table,
)

# revision identifiers, used by Alembic.
revision = "roc_dingo_0014_add_solohk"
down_revision = "roc_dingo_0013_fix_db"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    create_table(
        "solo_hk_param",
        sa.Column("id", sa.BIGINT(), nullable=False),
        sa.Column("name", sa.String(length=16), nullable=False),
        sa.Column("utc_time", postgresql.TIMESTAMP(), nullable=False),
        sa.Column("unit", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("eng_value", sa.String(), nullable=True),
        sa.Column("raw_value", sa.String(), nullable=True),
        sa.Column("sha", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sha"),
        schema="pipeline",
    )
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    execute(grant_user_seq)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    drop_table("solo_hk_param", schema="pipeline")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
"""Modify the view file_log_per_date
 - exclude CAL files from the count
 - try to guess the date with the filename if needed

Revision ID: roc_dingo_0018_modify_views_per_date
Revises: roc_dingo_0017_add_efecsevents
Create Date: 2021-11-22 16:59:56.553350

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "roc_dingo_0019_modify_views_per_date"
down_revision = "roc_dingo_0018b_add_product_to_file_log"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("COMMIT")

    # ensure the date is correctly guessed
    # skip SK files from this monitoring as we do not produce them
    op.execute("""CREATE OR REPLACE VIEW pipeline.file_log_per_date AS
        select coalesce(
                date(f1.start_time),
                date(f1.end_time),
                case when f1.basename ~ '^.*_[0-9]{8}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_([0-9]{4})([0-9]{2})([0-9]{2})_V.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_[0-9]{8}T[0-9]{6}\-[0-9]{8}T[0-9]{6}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})T[0-9]{6}\-.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_V[0-9]{14}\.xml$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})[0-9]{6}\.xml$',
                        '\\1-\\2-\\3') as date)
                else null
                end
            ) as date,
            count(f1.state='OK' or null) as OK,
            count(f1.state='ERROR' or null) as ERROR,
            count(f1.state='WARNING' or null) as WARNING,
            count(*)
        from pipeline.file_log f1
        where f1.level NOT IN ( 'CAL', 'SK')
        and f1.is_latest and not f1.is_removed
        group by coalesce(
                date(f1.start_time),
                date(f1.end_time),
                case when f1.basename ~ '^.*_[0-9]{8}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_([0-9]{4})([0-9]{2})([0-9]{2})_V.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_[0-9]{8}T[0-9]{6}\-[0-9]{8}T[0-9]{6}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})T[0-9]{6}\-.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_V[0-9]{14}\.xml$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})[0-9]{6}\.xml$',
                        '\\1-\\2-\\3') as date)
                else null
                end
            )
        order by date
    """)

    # useless now (SK files skipped by default)
    op.execute("DROP VIEW IF EXISTS pipeline.file_log_per_date_wo_sk")

    # monitor files w/o warnings for missing SPICE_KERNELS attribute
    op.execute("""CREATE OR REPLACE VIEW pipeline.file_log_per_date_wo_sk_warn AS
        select coalesce(
                date(f1.start_time),
                date(f1.end_time),
                case when f1.basename ~ '^.*_[0-9]{8}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_([0-9]{4})([0-9]{2})([0-9]{2})_V.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_[0-9]{8}T[0-9]{6}\-[0-9]{8}T[0-9]{6}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})T[0-9]{6}\-.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_V[0-9]{14}\.xml$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})[0-9]{6}\.xml$',
                        '\\1-\\2-\\3') as date)
                else null
                end
            ) as date,
            count(f1.state='OK' or null) as OK,
            count(f1.state='ERROR' or null) as ERROR,
            count(f1.state='WARNING' or null) as WARNING,
            count(*)
        from pipeline.file_log f1
        where f1.level NOT IN ( 'CAL', 'SK')
        and f1.error_log != '''SPICE_KERNELS: NO_SUCH_ATTR: Named attribute not found in this CDF.'''
        and f1.is_latest and not f1.is_removed
        group by coalesce(
                date(f1.start_time),
                date(f1.end_time),
                case when f1.basename ~ '^.*_[0-9]{8}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_([0-9]{4})([0-9]{2})([0-9]{2})_V.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_[0-9]{8}T[0-9]{6}\-[0-9]{8}T[0-9]{6}_V.*$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})T[0-9]{6}\-.*$',
                        '\\1-\\2-\\3') as date)
                when f1.basename ~ '^.*_V[0-9]{14}\.xml$' then
                    cast(regexp_replace(
                        f1.basename,
                        '^.*_V([0-9]{4})([0-9]{2})([0-9]{2})[0-9]{6}\.xml$',
                        '\\1-\\2-\\3') as date)
                else null
                end
            )
        order by date
    """)


def downgrade():
    op.execute("COMMIT")
    op.execute("""CREATE OR REPLACE VIEW pipeline.file_log_per_date AS
        select date(start_time) as date,
            count(state='OK' or null) as OK,
            count(state='ERROR' or null) as ERROR,
            count(state='WARNING' or null) as WARNING,
            count(*)
        from pipeline.file_log f1
        where f1.start_time is not null
        and f1.is_latest and not f1.is_removed
        group by (date(start_time))
        order by date
    """)

    op.execute("""CREATE OR REPLACE VIEW pipeline.file_log_per_date_wo_sk AS
        select date(start_time) as date,
            count(state='OK' or null) as OK,
            count(state='ERROR' or null) as ERROR,
            count(state='WARNING' or null) as WARNING,
            count(*)
        from pipeline.file_log f1
        where f1.start_time is not null
        and level != 'SK'
        and f1.is_latest and not f1.is_removed
        group by (date(start_time))
        order by date
    """)

    op.execute("DROP VIEW IF EXISTS pipeline.file_log_per_date_wo_sk_warn")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Functions used in all DINGO tests
"""

import os
import copy
from pathlib import Path
import argparse
from typing import Union

import yaml

from sqlalchemy import create_engine, inspect
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import Session
from sqlalchemy.dialects import postgresql

from poppy.core.logger import logger
from poppy.core.test import CommandTestCase
from poppy.core.configuration import Configuration

from roc.dingo.models.file import FileLog
from roc.dingo.constants import TEST_DATABASE
from roc.dingo.constants import DINGO_CACHE_DIR


# Models to be adapted for each Dingo model
#
# from poppy.core.test import TaskTestCase, CommandTestCase
# class DingoTaskTestCase(TaskTestCase):
#
#     @classmethod
#     def setup_class(cls):
#         logger.debug('setup_class DingoLogFileTaskTestCase')
#         d = DingoTest()
#         d.get_test_data()
#
#    #     # --- database setup ---
#    #     session = DingoTestLogFile.setup_session()
#
#    # @classmethod
#    # def teardown_class(cls):
#    #     logger.debug('teardown_class DingoTaskTestCase')
#
#    #     # --- close session
#    #     session.close()
#
# class DingoCommandTestCase(CommandTestCase):
#
#     @classmethod
#     def setup_class(cls):
#         logger.debug('setup_class DingoCommandTestCase')
#         d = DingoTest()
#         d.get_test_data()


class DingoTest:
    # base_url = (
    #     "https://rpw.lesia.obspm.fr/roc/data/private/devtest/roc/test_data/rodp/dingo"  # noqa: E501
    # )
    # base_path = (
    #     "/volumes/plasma/rpw/roc/data/https/private/devtest/roc/test_data/rodp/dingo"  # noqa: E501
    # )
    #
    # # test credentials
    # host = "roc2-dev.obspm.fr"
    # username = os.environ.get("ROC_TEST_USER", "roctest")
    # password = None

    def __init__(self):
        logger.debug("DingoTest setup_class()")
        # logger.debug(f"base_url = {self.base_url}")
        # logger.debug(f"base_path = {self.base_path}")
        # try:
        #    self.password = os.environ["ROC_TEST_PASSWORD"]
        # except KeyError:
        #    raise KeyError(
        #        "You have to define the test user password using"
        #        'the "ROC_TEST_PASSWORD" environment variable'
        #    )

    @staticmethod
    def compare_data_to_expected(
        expected_data: dict, data: dict, ignore: Union[list, None] = None
    ):
        """
        Compare expected_data to the data array dictionary
            -> check if every expected item is in data
            -> check if every data item is in expected data

            :param expected_data: expected data
            :type expected_data: dict
            :param data: the data to be compared
            :type data: dict
            :param ignore: items keys to ignore
            :type ignore: list
        """
        if ignore is None:
            ignore = []

        # check length
        assert len(expected_data) == len(data)

        # check content
        for item in data:
            assert item in expected_data, "item {} is not in expected data".format(item)
            for key in ignore:
                data[item].pop(key, None)
                expected_data[item].pop(key, None)

        for item in expected_data:
            assert item in data, "item {} is not in result".format(item)
            assert data[item] == expected_data[item], (
                "item {} is not the one expected :expected {} / got {}".format(
                    item, expected_data[item], data[item]
                )
            )

    @staticmethod
    def read_data(file: str) -> list:
        """
        Read a file located in the data-tasks directory
        and returns file content as an array of strings

        :param file: file to open
        :type file: str
        :return: data inside file
        rtype: list
        """
        path = os.path.abspath(os.path.dirname(__file__))
        path = os.path.join(path, "data-tasks", file)
        data = []
        with open(path, "r") as f:
            # read a line corresponding to a directory
            for line in f.readlines():
                # append the data
                data.append(line.strip())
        return data

    @staticmethod
    def get_test_data_path() -> str:
        """
        Read the config file and returns ROC_TEST_DATA_PATH
        which is the path where to store the test dataset locally

        :return: path to the test data directory
        :rtype: str
        """
        # Define default value
        data_test_path = DINGO_CACHE_DIR

        conf = DingoTest.load_configuration()

        # Check if ROC_TEST_DATA_PATH env. variable is defined
        # in: (1) config file, (2) shell env.
        for source in [conf["environment"], os.environ]:
            try:
                data_test_path = source["ROC_TEST_DATA_PATH"]
            except KeyError:
                # logger.debug('Env. variable ROC_TEST_DATA_PATH not set')
                pass
            else:
                break

        logger.debug(f"ROC_TEST_DATA_PATH = {data_test_path}")
        return data_test_path

    # def get_test_data(self):
    #     """
    #     Get the test dataset indicated by the environment variable
    #     ROC_TEST_DATA_PATH
    #
    #     Try to make a rsync with the roctest account
    #     A public kay has to be setup on the server to allow connexion
    #     If the command is not available (Windows),
    #     use the download_file() method
    #
    #     """
    #     data_test_path = DingoTest.get_test_data_path()
    #     os.makedirs(data_test_path, exist_ok=True)
    #
    #     try:
    #         logger.info("Starting rsync")
    #         ssh_option = "\"ssh -o 'StrictHostKeyChecking no'\""
    #         rsync_cmd = "rsync -e {} -irtzuv {}@{}:{}/ {}/".format(
    #             ssh_option, self.username, self.host, self.base_path, data_test_path
    #         )
    #         logger.info("Executing " + rsync_cmd)
    #         output = os.popen(rsync_cmd)
    #         rsync_output = output.read()
    #         if output.close() is not None:
    #             raise ValueError("Rsync failed : {}".format(rsync_output))
    #     except ValueError:
    #         logger.info("Rsync failed, using download_test_data()")
    #         self.download_test_data(data_test_path)

    # def download_test_data(self, data_test_path):
    #     """
    #     Download the manifest.txt file located at self.base_url
    #     And for each file, download it only if the file does not exist
    #     in data_test_path
    #     """
    #     logger.debug("download_test_data()")
    #     manifest_filepath = osp.join(data_test_path, "manifest.txt")
    #
    #     manifest_file_url = self.base_url + "/manifest.txt"
    #     auth = (self.username, self.password)
    #
    #     file_list = list(
    #         self.load_manifest_file(manifest_filepath, manifest_file_url, auth=auth)
    #     )
    #
    #     for relative_filepath in file_list:
    #         # skip empty strings
    #         if not relative_filepath:
    #             continue
    #
    #         # get the complete filepath
    #         filepath = osp.join(data_test_path, relative_filepath)
    #         os.makedirs(osp.dirname(filepath), exist_ok=True)
    #
    #         # download it only if it does not exist
    #         if not osp.isfile(filepath):
    #             logger.info("Downloading {}".format(filepath))
    #             download_file(
    #                 filepath, f"{self.base_url}/{relative_filepath}", auth=auth
    #             )
    #
    # @staticmethod
    # def load_manifest_file(manifest_filepath, manifest_file_url, auth=None):
    #     """
    #     Read the manifest.txt file located at manifest_file_url
    #     and returns the list composed by the file list
    #     """
    #
    #     download_file(manifest_filepath, manifest_file_url, auth=auth)
    #
    #     with open(manifest_filepath) as manifest_file:
    #         for line in manifest_file:
    #             yield line.strip("\n\r")
    #
    #     os.remove(manifest_filepath)

    @staticmethod
    def load_configuration(config_file: Path = None) -> Configuration:
        """
        Load pipeline configuration file in JSON format.

        :param config_file: Path of the configuration file.
        If not passed, try to retrieve file path from
        PIPELINE_CONFIG_FILE environment variable
        :type config_file: Path
        :return: pipeline configuration data
        :rtype: poppy.core.configuration.Configuration
        """
        if config_file:
            configuration = Configuration(str(config_file))
        else:
            configuration = Configuration(os.getenv("PIPELINE_CONFIG_FILE", None))
        configuration.read()

        return configuration

    @staticmethod
    def get_spice_kernel_dir():
        """
        Returns SPICE kernels directory

        :return: spice_kernels_dir
        """
        # Define default value
        spice_kernels_dir = os.path.join(Path.cwd(), "data", "spice_kernels")

        # Get pipeline configuration parameters
        conf = DingoTest.load_configuration()

        # Check if SPICE_KERNEL_PATH env. variable is defined
        # in: (1) config file, (2) shell env.
        for source in [conf["environment"], os.environ]:
            try:
                spice_kernels_dir = os.path.join(source["SPICE_KERNEL_PATH"])
            except Exception as e:
                logger.debug(e)
                # logger.debug('Env. variable SPICE_KERNEL_PATH not set')
                pass
            else:
                break

        return spice_kernels_dir

    @staticmethod
    def setup_session(config_file: Path = None) -> Session:
        """
        Set up database configuration.

        :param config_file: Pipeline/database configuration file path
        :type config_file: Path
        :return: database session
        :rtype: Session
        """
        # Read config file
        conf = DingoTest.load_configuration(config_file=config_file)

        database_info = list(
            filter(
                lambda db: db["identifier"] == TEST_DATABASE, conf["pipeline.databases"]
            )
        )[0]

        # Create an Engine, which the Session will use for connection resources
        engine = create_engine(
            "{}://{}@{}/{}".format(
                database_info["login_info"]["vendor"],
                database_info["login_info"]["user"],
                database_info["login_info"]["address"],
                database_info["login_info"]["database"],
            )
        )
        # create a configured "Session" class
        Session = sessionmaker(bind=engine, autocommit=False)
        # create a Session
        session = Session()

        return session

    @staticmethod
    def get_db_values(item):
        """
        returns a dict for any object in DB returned by session.query

        :param item: the object
        """

        return copy.copy(item.__dict__)

    @staticmethod
    def get_db_values_columns_only(item):
        """
        Convert an input item as returned by SQLAlchemy.query into dictionary.
        (Same than DingoTest.get_db_values(), but with columns only)
        See https://stackoverflow.com/questions/1958219/how-to-convert-sqlalchemy-row-object-to-a-python-dict

        :param item: SQLAlchemy table model class instance
        :return: dictionary with columns as keywords and entries as values
        """
        return {
            c.expression.name: getattr(item, c.key)
            for c in inspect(item).mapper.column_attrs
        }

    @staticmethod
    def clear_db(command_test_case: CommandTestCase, log_level: str = "INFO"):
        # Clear database
        clear_db_cmd = ["pop", "db", "downgrade", "base", "-ll", log_level]
        command_test_case.run_command(clear_db_cmd)

    @staticmethod
    def create_db(command_test_case: CommandTestCase, log_level: str = "INFO"):
        # Run database migrations
        create_db_cmd = ["pop", "db", "upgrade", "heads", "-ll", log_level]
        command_test_case.run_command(create_db_cmd)

    @staticmethod
    def load_idb(
        command_test_case: CommandTestCase,
        idb_source: str,
        idb_version: str,
        install_dir: str = None,
        log_level="INFO",
    ) -> str:
        if install_dir is None:
            install_dir = DINGO_CACHE_DIR
        idb_dir = os.path.join(install_dir, f"idb-{idb_source}-{idb_version}")
        os.makedirs(idb_dir, exist_ok=True)

        # IDB loading
        load_idb_cmd = [
            "pop",
            "idb",
            "install",
            "--force",
            "--install-dir",
            idb_dir,
            "-s",
            idb_source,
            "-v",
            idb_version,
            "--load",
            "-ll",
            log_level,
        ]
        command_test_case.run_command(load_idb_cmd)

        return idb_dir

    @staticmethod
    def filelog_to_yml(yml_outfile: Path, level: list = None, config_file: Path = None):
        """
        Export pipeline.file_log database table content
        into an output yaml format file.
        Can be used to quickly generated the "roc_data_files_all_dump.yml

        :param yml_outfile: Path and name of the output yaml format file
        :type yml_outfile: Path
        :param level: Filter rows by data processing level
        :type level: list
        :param config_file: Path to configuration file
        :type config_file: Path
        :return: None
        """
        if not level:
            level = ["L0", "L1", "HK", "L1R", "L2", "L3", "CAL", "BIA"]

        session = DingoTest.setup_session(config_file=config_file)
        query = session.query(FileLog).filter(FileLog.level.in_(level))
        try:
            results = query.all()
        except NoResultFound as e:
            logger.error(f"No entry found in pipeline.file_log with {level}:\n{e}")
            logger.debug(
                str(
                    query.statement.compile(
                        dialect=postgresql.dialect(),
                        compile_kwargs={"literal_binds": True},
                    )
                )
            )
        else:
            output_data = {}
            for i, row in enumerate(results):
                row_dict = row.as_dict()
                output_data[row_dict["basename"]] = {k: v for k, v in row_dict.items()}

            with open(str(yml_outfile), "w") as f:
                yaml.dump(output_data, f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--yml_file", nargs=1, type=Path, help="Path and name of output yaml file"
    )
    parser.add_argument(
        "-l",
        "--level",
        nargs="+",
        type=str,
        help="Filter rows by data processing level",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        nargs=1,
        type=Path,
        help="Path and name of config JSON file",
    )
    args = parser.parse_args()

    DingoTest.filelog_to_yml(args.yml_file[0], level=args.level)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import uuid
from datetime import datetime

import h5py
from sqlalchemy import and_
import pandas as pd
import numpy as np

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import FileTarget

from roc.rpl.packet_structure.data import Data

from roc.dingo.models.data import HfrTimeLog
from roc.dingo.tasks import get_l0_files, load_param
from roc.dingo.tools import (
    glob_paths,
    valid_time,
    query_db,
    get_columns,
    load_spice,
    insert_in_db,
)
from roc.dingo.constants import (
    PIPELINE_DATABASE,
    TIME_DAILY_STRFORMAT,
    TRYOUTS,
    TIME_WAIT_SEC,
    SQL_LIMIT,
    TIME_ISO_STRFORMAT,
    HFR_SCIENCE_PACKETS,
)

__all__ = ["L0ToHfrTimeLog"]


class L0ToHfrTimeLog(Task):
    """
    Insert HFR data in input L0 files into the
    in pipeline.hfr_time_log table in ROC database
    See https://gitlab.obspm.fr/ROC/RCS/THR_CALBAR/-/issues/76 for details
    about why this task is needed
    """

    plugin_name = "roc.dingo"
    name = "l0_to_hfrtimelog"

    def add_targets(self):
        self.add_input(
            identifier="rpw_l0_files",
            many=True,
            filepath=get_l0_files,
            target_class=FileTarget,
        )

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):
        # get the input l0 files
        self.l0_files = glob_paths(self.inputs["rpw_l0_files"].filepath)

        # Get start_time input value
        self.start_time = valid_time(
            self.pipeline.get("start_time", default=[None])[0],
            str_format=TIME_DAILY_STRFORMAT,
        )

        # Get end_time input value
        self.end_time = valid_time(
            self.pipeline.get("end_time", default=[None])[0],
            str_format=TIME_DAILY_STRFORMAT,
        )

        # get a database session, table model and table columns (except primary key)
        self.session = Connector.manager[PIPELINE_DATABASE].session
        self.model = HfrTimeLog
        self.columns = get_columns(self.model, remove=["id"])

        # Get tryouts from pipeline properties
        self.tryouts = self.pipeline.get("tryouts", default=[TRYOUTS], create=True)[0]

        # Get wait from pipeline properties
        self.wait = self.pipeline.get("wait", default=[TIME_WAIT_SEC], create=True)[0]

        # Retrieve --limit keyword value
        self.limit = self.pipeline.get(
            "limit",
            default=[SQL_LIMIT],
        )[0]

        # Get or create failed_files list from pipeline properties
        self.failed_files = self.pipeline.get("failed_files", default=[], create=True)

        # Get or create processed_files list from pipeline properties
        self.processed_files = self.pipeline.get(
            "processed_files", default=[], create=True
        )

        # Load spice_manager class
        self.spice = load_spice()

        # Define query filters for existing data in database
        self.filters = []
        if self.start_time:
            self.filters.append(self.model.acq_time >= str(self.start_time))
        if self.end_time:
            self.filters.append(self.model.acq_time < str(self.end_time))

    def run(self):
        # Define task job ID (long and short)
        self.job_uuid = str(uuid.uuid4())
        self.job_id = f"L0ToHfrTimeLog-{self.job_uuid[:8]}"
        logger.info(f"Task {self.job_id} is starting")
        try:
            self.setup_inputs()
        except Exception as e:
            logger.error(f"Initializing inputs has failed for {self.job_id}!")
            logger.debug(e)
            self.pipeline.exit()
            return

        n_l0_files = len(self.l0_files)
        logger.info(f"{n_l0_files} RPW L0 files to process")
        if n_l0_files == 0:
            return

        # Loops over each input RPW L0 file
        inserted_count = 0
        updated_count = 0
        failed_count = 0
        for i, current_l0_file in enumerate(self.l0_files):
            logger.info(
                f"Processing {current_l0_file}    ({n_l0_files - i - 1} remaining)"
            )

            try:
                with h5py.File(current_l0_file, "r") as l0:
                    # Check if the SCLK SPICE kernel used to compute TM utc times in the input L0 file
                    # is older or newer than the l0 file date (if kernel is older, then it means that utc times are computed
                    # with predictive time coefficients. In this case, the utc times are not stored in the database
                    # in order to avoid confusion)
                    current_l0_datetime = datetime.strptime(
                        l0.attrs["Datetime"][:8],
                        TIME_DAILY_STRFORMAT,
                    )
                    # Skip file if outside of the [start_time, end_time] time range
                    # (if any)
                    if (
                        self.start_time
                        and self.start_time.date() > current_l0_datetime.date()
                    ):
                        logger.info(
                            f"{current_l0_file} older than {self.start_time.date()}, skip it"
                        )
                        continue
                    if (
                        self.end_time
                        and self.end_time.date() < current_l0_datetime.date()
                    ):
                        logger.info(
                            f"{current_l0_file} newer than {self.end_time.date()}, skip it"
                        )
                        continue

                    # Retrieving L0 start/end times
                    self.l0_start_time = datetime.strptime(
                        l0.attrs["TIME_MIN"], TIME_ISO_STRFORMAT
                    )
                    self.l0_end_time = datetime.strptime(
                        l0.attrs["TIME_MAX"], TIME_ISO_STRFORMAT
                    )

                    # Retrieve existing data from pipeline.hfr_time_log table before each L0 file processing
                    # (Database content may change between two L0 file processes)
                    logger.debug(
                        f"Getting existing hfr_time_log data between {self.start_time} and {self.end_time}"
                    )
                    # Return existing data as a pandas.DataFrame object
                    table_data = query_db(
                        self.session,
                        self.model,
                        filters=(and_(*self.filters) if self.filters else None),
                        tryouts=self.tryouts,
                        wait=self.wait,
                        limit=self.limit,
                    )
                    n_existing = table_data.shape[0]
                    # If data have been found in the database, then building and adding a new "coarse_fine" variable.
                    # This variable will be used to handle and compare times in the data
                    # (more simple than with [coarse, fine] vectors)
                    if n_existing > 0:
                        table_data["coarse_fine"] = table_data.apply(
                            lambda x: self.merge_coarse_fine(
                                x["coarse_time"], x["fine_time"]
                            ),
                            axis=1,
                        )
                    else:
                        logger.info(
                            f"No existing hfr_time_log data found in database between {self.start_time} and {self.end_time}"
                        )

                    # Loop over TM packets in L0 file
                    for packet_name in l0["TM"].keys():
                        # Only process science TM packets from HFR
                        if packet_name not in HFR_SCIENCE_PACKETS:
                            continue

                        # Get packet header
                        packet_header = load_param(
                            l0["TM"][packet_name]["packet_header"]
                        )
                        # Get packet data field header
                        data_field_header = load_param(
                            l0["TM"][packet_name]["data_field_header"]
                        )
                        # Get packet data
                        packet_data = load_param(l0["TM"][packet_name]["source_data"])

                        # Concatenate packet headers and data dataframes
                        packet_df = pd.concat(
                            [packet_header, data_field_header, packet_data], axis=1
                        )

                        # Add packet creation UTC time
                        packet_df["utc_time"] = l0["TM"][packet_name]["utc_time"][()]

                        # Get number of packets
                        n_packet = packet_df.shape[0]

                        logger.info(
                            f"{n_packet} {packet_name} packets retrieved from {current_l0_file}"
                        )

                        # Get mode
                        if "NORMAL" in packet_name:
                            self.mode = 0
                        elif "BURST" in packet_name:
                            self.mode = 1
                        elif "CALIBRATION" in packet_name:
                            self.mode = 2
                        else:
                            raise ValueError(
                                f"Unknown HFR mode for {packet_name} in {current_l0_file}"
                            )

                        # Convert PA_THR_ACQUISITION_TIME values into unique integers
                        # (To be comparable with "coarse_fine" variable in database table_data)
                        packet_df["coarse_fine"] = packet_df[
                            "PA_THR_ACQUISITION_TIME"
                        ].apply(lambda x: self.merge_coarse_fine(x[0], x[1]))

                        # Do the same for packet creation time
                        packet_df["packet_time"] = packet_df["time"].apply(
                            lambda x: self.merge_coarse_fine(x[0], x[1])
                        )

                        # Extract values of delta times for HF1 and HF2 for each packet
                        # Only for NORMAL and BURST packets
                        if self.mode != 2:
                            try:
                                packet_df["delta_time1"], packet_df["delta_time2"] = (
                                    zip(
                                        *packet_df.apply(self.extract_deltatime, axis=1)
                                    )
                                )
                            except Exception as e:
                                logger.error(
                                    f"Extraction delta_time values from {packet_name} in {current_l0_file} has failed!"
                                )
                                logger.debug(e)
                                self.failed_files.append(current_l0_file)
                                break

                        # Get list of unique PA_THR_ACQUISITION_TIME values
                        unique_coarse_fine = packet_df["coarse_fine"].unique()
                        n_entry = unique_coarse_fine.shape[0]
                        logger.info(
                            f"{n_entry} {packet_name} entries to insert into hfr_time_log table for {current_l0_file}"
                        )
                        for i in range(n_entry):
                            # Get subset of packet data for current acquisition time
                            current_indices = packet_df.index[
                                unique_coarse_fine[i] == packet_df["coarse_fine"]
                            ]
                            current_df = packet_df.iloc[current_indices].reset_index(
                                drop=True
                            )

                            # Fill new database entry dictionary
                            current_entry = {}
                            current_entry["coarse_time"] = int(
                                current_df["PA_THR_ACQUISITION_TIME"][0][0]
                            )
                            current_entry["fine_time"] = int(
                                current_df["PA_THR_ACQUISITION_TIME"][0][1]
                            )
                            current_entry["acq_time"] = self.spice.cuc2datetime(
                                current_entry["coarse_time"], current_entry["fine_time"]
                            )
                            current_entry["mode"] = self.mode

                            if self.mode != 2:
                                # For delta_time1 and delta_time2 columns, make sure to concatenate all values
                                # Save delta_time1 and delta_time2 as (JSONB format) dictionaries
                                # with packet creation times as keywords
                                # and delta_time1 and delta_time2 block values as values
                                current_entry["delta_time1"] = {}
                                current_entry["delta_time2"] = {}
                                for j in range(current_df.shape[0]):
                                    current_entry["delta_time1"][
                                        str(int(current_df["packet_time"][j]))
                                    ] = (
                                        current_df["delta_time1"][j]
                                        .astype(int)
                                        .tolist()
                                    )
                                    current_entry["delta_time2"][
                                        str(int(current_df["packet_time"][j]))
                                    ] = (
                                        current_df["delta_time2"][j]
                                        .astype(int)
                                        .tolist()
                                    )

                            # Check if data for current acquisition time already in the database
                            try:
                                current_table_df = table_data[
                                    unique_coarse_fine[i] == table_data["coarse_fine"]
                                ]
                                current_table_df.reset_index(inplace=True)
                            except Exception as e:
                                logger.debug(
                                    f"No {packet_name} entry found in the hfr_time_log table for {current_entry['acq_time']}:\n{e}"
                                )
                                data_to_insert = current_entry
                            else:
                                # if data found, then merge data before inserting
                                data_to_insert = self.merge_data(
                                    current_entry, current_table_df
                                )

                                if data_to_insert is None:
                                    logger.info(
                                        f"No new {packet_name} data to insert in hfr_time_log table for {current_entry['acq_time']}"
                                    )
                                    continue
                                elif self.mode != 2:
                                    # Sort delta_time dictionaries by ascending packet creation times
                                    data_to_insert["delta_time1"] = dict(
                                        sorted(data_to_insert["delta_time1"].items())
                                    )
                                    data_to_insert["delta_time2"] = dict(
                                        sorted(data_to_insert["delta_time2"].items())
                                    )

                            # define columns to check for update
                            update_fields_kwargs = {
                                key: val
                                for key, val in data_to_insert.items()
                                if key not in ["delta_time1", "delta_time2"]
                            }

                            # Insert entry in the hfr_time_log table
                            insert_status = insert_in_db(
                                self.session,
                                self.model,
                                data_to_insert,
                                update_fields=data_to_insert,
                                update_fields_kwargs=update_fields_kwargs,
                                tryouts=self.tryouts,
                                wait=self.wait,
                            )
                            if insert_status < 0:
                                logger.error(
                                    f"Inserting {update_fields_kwargs} in pipeline.hfr_time_log table "
                                    f"has failed for task {self.job_id}"
                                )
                                failed_count += 1
                            elif insert_status > 1:
                                logger.info(
                                    f"{update_fields_kwargs} updated in hfr_time_log"
                                )
                                updated_count += 1
                            else:
                                logger.debug(
                                    f"{update_fields_kwargs} inserted into hfr_time_log table for task {self.job_id}"
                                )
                                inserted_count += 1

            except Exception as e:
                self.failed_files.append(current_l0_file)
                logger.error(f"Cannot process {current_l0_file} in task {self.job_id}!")
                logger.debug(e)
                break

        logger.info(f"{inserted_count} new entries inserted into hfr_time_log table")
        if updated_count > 0:
            logger.info(f"{updated_count} entries updated in hfr_time_log table")
        if failed_count > 0:
            logger.error(
                f"hfr_time_log table insertion has failed for {failed_count} entries"
            )

        logger.info(f"Task {self.job_id} has ended")

    def merge_data(self, current_entry, current_table_df):
        """
        Merge properly l0 file data for current acquisition time
        with data already found in the hfr_time_log table.

        :param current_entry: dictionary with entry to insert for current acq. time
        :param current_table_df: pandas DataFrame with existing
                                 hfr_time_log table data for current acq. time
        :return: new dictionary with merged data. NoneType if no new data to merge
        """
        # No merge for CALIBRATION mode
        if current_entry["mode"] == 2:
            return current_entry

        # if no data in database for current entry, then return as is
        if current_table_df.shape[0] == 0:
            logger.debug(
                f"No delta time values to merge for {current_entry['acq_time']}"
            )
            return current_entry

        # Get delta_time1 and delta_time2 from table data
        table_delta_time1 = current_table_df["delta_time1"][0]
        table_delta_time2 = current_table_df["delta_time2"][0]
        # Check if current_entry already saved in database
        if all(
            k in table_delta_time1 for k in current_entry["delta_time1"].keys()
        ) and all(k in table_delta_time2 for k in current_entry["delta_time2"].keys()):
            # If already saved, then return None
            return None

        # Else merge delta_time dictionaries from database and l0 file for current acquisition time
        merged_delta_time1 = {**table_delta_time1, **current_entry["delta_time1"]}
        merged_delta_time2 = {**table_delta_time2, **current_entry["delta_time2"]}

        # Update current entry with new delta times values
        current_entry["delta_time1"] = merged_delta_time1
        current_entry["delta_time2"] = merged_delta_time2

        return current_entry

    def merge_coarse_fine(self, coarse, fine):
        """
        return input coarse and fine parts of CCSDS CUC time
        as an unique integer

        :param coarse: coarse part of CUC time
        :param fine: fine part of CUC time
        :return: resulting unique integer
        """
        return int(coarse * 100000 + fine)

    def extract_deltatime(self, packet_data):
        """
        Extract HF1 / HF2 delta time values from input HFR binary data blocks.
        Only works with TM_THR_SCIENCE_NORMAL_HFR and TM_THR_SCIENCE_BURST_HFR packets

        :param packet_data: pandas.DataFrame storing packet data
        :return: tuple containing delta time values for HF1 and HF2 bands
        """
        # size of the number of unsigned int
        size = packet_data["PA_THR_HFR_DATA_CNT"]

        # Define block counter parameter name
        # (depends on the mode)
        if self.mode == 0:
            cnt_param = "PA_THR_N_BLOCK_CNT"
        else:
            cnt_param = "PA_THR_B_BLOCK_CNT"

        # number of blocks
        nblocks = packet_data[cnt_param]

        # transform the data into byte array
        # store because of the garbage collector removing reference while
        # processing in cython
        byte = (
            np.array(packet_data["PA_THR_HFR_DATA"][:size], dtype=np.uint16)
            .byteswap()
            .newbyteorder()
            .tobytes()
        )
        data = Data(byte, len(byte))

        # init offset
        offset = 0

        # init output values
        delta_time1 = np.empty(nblocks, dtype=np.uint32)
        delta_time2 = np.empty(nblocks, dtype=np.uint32)

        # loop over blocks
        for j in range(nblocks):
            # Skip command control, rpw status and temperatures bits
            offset += 8

            # read hfr setup
            hfr_setup_cmd = self._hfr_setup(data, offset)
            offset += 2

            # Skip input setup bits
            offset += 2

            # read hfr sweep setup
            hfr_sweep_cmd = self._hfr_sweep_setup(data, offset)
            n1 = hfr_sweep_cmd[2]
            n2 = hfr_sweep_cmd[0]
            offset += 4

            # read do analysis command
            offset += 1
            do_analysis_cmd = self._do_analysis(data, offset)
            offset += 1

            # read delta times
            deltas = self._delta_times(data, offset)
            delta_time1[j] = deltas[0]
            delta_time2[j] = deltas[1]
            offset += 8

            # initialize bit count
            bit_count = 0

            # Count bands size
            bit_count = self._get_band_size(
                do_analysis_cmd[4],
                hfr_setup_cmd[3],
                hfr_setup_cmd[2],
                n1,
                n2,
                bit_count,
            )
            bit_count = self._get_band_size(
                do_analysis_cmd[3],
                hfr_setup_cmd[3],
                hfr_setup_cmd[2],
                n1,
                n2,
                bit_count,
            )

            # compute the new offset for the block
            offset += bit_count // 8

            # align
            if offset % 2 != 0:
                offset += 1

        return delta_time1, delta_time2

    def _delta_times(self, data, start):
        """
        Read HF1 and HF2 delta times
        from TM block headers.
        """

        delta1 = data.u32p(start, 0, 32)
        delta2 = data.u32p(start + 4, 0, 32)

        return delta1, delta2

    def _get_band_size(self, ch, band1, band2, n1, n2, bit_count):
        """
        Get HFR bands total bit size.
        """

        # if channel is activated
        if ch == 1:
            # read the first band
            if band1 == 1:
                bit_count += 12 * n1

            # second band
            if band2 == 1:
                bit_count += 12 * n2

        return bit_count

    def _hfr_setup(self, data, start):
        """
        Return tuple of values from the HFR setup command.
        """
        tmp = start + 1

        # read parameters
        av = data.u8p(tmp, 6, 2)
        hf1 = data.u8p(tmp, 5, 1)
        hf2 = data.u8p(tmp, 4, 1)
        initial_frequency = data.u16p(start, 3, 9)
        sw = data.u8p(start, 2, 1)

        return sw, initial_frequency, hf2, hf1, av

    def _hfr_sweep_setup(self, data, start):
        """
        Return tuple of values from the HFR sweep setup command.
        """

        # read parameters
        hf1_size = data.u8p(start + 3, 4, 4)
        hf1_number = data.u16p(start + 2, 3, 9)
        hf2_size = data.u8p(start + 1, 7, 4)
        hf2_number = data.u16p(start, 6, 9)

        return hf2_number, hf2_size, hf1_number, hf1_size

    def _do_analysis(self, data, start):
        """
        Return tuple of values from the DoAnalysis command.
        """

        # read parameters
        ch1 = data.u8p(start, 7, 1)
        ch2 = data.u8p(start, 6, 1)
        eos = data.u8p(start, 5, 1)
        mod = data.u8p(start, 4, 1)
        it = data.u8p(start, 0, 4)

        return it, mod, eos, ch2, ch1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo efecs_to_db command and related tasks.
"""

import glob
import os
import shlex
import tarfile

import unittest.mock as mock
from pathlib import Path

from poppy.core.conf import Settings
from poppy.core.logger import logger
from poppy.core.generic.requests import download_file
from poppy.core.test import CommandTestCase

from roc.dingo.models.data import EfecsEvents

# Tests on roc.dingo efecs_to_db command
from roc.dingo.tests.test_dingo import DingoTest
from roc.dingo.constants import DINGO_CACHE_DIR


class TestDingoEfecsToDbCommand(CommandTestCase):
    def setup_method(self, method):
        super().setup_method(method)

        self.session.query(EfecsEvents).delete()
        self.session.flush()

    def teardown_method(self, method):
        super().teardown_method(method)

    def test_efecs_to_db(self):
        # Name of the command to test
        cmd = "efecs_to_db"

        # First get test data to insert (i.e. EFECS XML input files)
        test_data_tarball = "test_data.tar.gz"

        self.install_dir = os.environ.get("ROC_TEST_DATA_PATH", DINGO_CACHE_DIR)
        test_data_dir = os.path.join(self.install_dir, cmd)
        Path(test_data_dir).mkdir(exist_ok=True, parents=True)
        test_data_path = os.path.join(test_data_dir, test_data_tarball)
        if not os.path.isfile(test_data_path):
            # Build test data tarball URL
            test_data_url = "/".join([DingoTest.base_url, cmd, test_data_tarball])
            # Get username/password
            test_dingo = DingoTest()

            download_file(
                test_data_path,
                test_data_url,
                auth=(test_dingo.username, test_dingo.password),
            )

        # Extract tarball
        if tarfile.is_tarfile(test_data_path):
            logger.info(f"Extracting {test_data_path} ...")
            with tarfile.open(test_data_path, "r:*") as tarball:
                tarball.extractall(path=test_data_dir, filter="fully_trusted")
        else:
            raise tarfile.ReadError(f"{test_data_path} is not a valid tarball!")

        # Get list of E-FECS files
        input_dir = os.path.join(test_data_dir, "inputs")
        input_files = sorted(list(glob.glob(os.path.join(input_dir, "EFECS_M??.xml"))))
        if not input_files:
            raise FileNotFoundError(f"No EFECS XML file found in {input_dir}")

        # initialize the main command
        command_to_test = " ".join(
            ["pop", "dingo", cmd, "--efecs-files", " ".join(input_files), "-ll", "INFO"]
        )

        # define the required plugins
        plugin_list = ["poppy.pop", "roc.dingo"]

        # run the command
        # force the value of the plugin list
        with mock.patch.object(
            Settings,
            "configure",
            autospec=True,
            side_effect=self.mock_configure_settings(
                dictionary={"PLUGINS": plugin_list}
            ),
        ):
            self.run_command(shlex.split(command_to_test))

        # Check that test data are in the database
        logger.debug("Querying efecs_events table in database ...")
        rows = self.session.query(EfecsEvents).all()

        assert isinstance(rows, list)
        assert len(rows) == 3378

        # Convert first/last row objects to dictionaries (for debugging)
        # first_row = DingoTest.get_db_values_columns_only(rows[0])
        # last_row = DingoTest.get_db_values_columns_only(rows[-1])
        # print(first_row)
        # print(last_row)

        assert rows[0].name == "ATT_DIST"
        assert rows[0].ltp_count == 1
        assert rows[-1].name == "NAV"
        assert rows[-1].ltp_count == 2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import os
import re

from datetime import datetime, timedelta

from sqlalchemy import and_
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import load_only

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import (
    PIPELINE_DATABASE,
    ROC_DATA_ROOT,
    DATA_ALLOWED_EXTENSIONS,
    SPICE_KERNEL_ALLOWED_EXTENSIONS,
    ROC_DATA_HTTPS_ROOT,
    ROC_SPICE_KERNEL_HTTPS_ROOT,
    OBSOLETE_ATTRS,
)
from roc.dingo.models.file import FileLog

from roc.dingo.constants import TIME_ISO_STRFORMAT

import h5py
from spacepy.pycdf import CDF


__all__ = ["LogFileToDb"]


class LogFileToDb(Task):
    """
    Parse ROC file tree and synchronize withe the ROC databaser
    """

    plugin_name = "roc.dingo"
    name = "log_file_to_db"
    files_to_update = []
    files_to_insert = []

    def add_targets(self):
        logger.debug("LogFileToDb() : add_targets")
        self.add_input(
            target_class=BaseTarget, many=True, identifier="roc_data_files_to_insert"
        )
        self.add_input(
            target_class=BaseTarget, many=True, identifier="roc_data_files_to_update"
        )

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug("LogFileToDb() : setup_inputs")

        # get the root file tree
        self.root = self.pipeline.get("root", default=ROC_DATA_ROOT, args=True)
        # ensure that there is / at the end
        self.root = os.path.join(self.root, "")

        # get files to update
        self.files_to_db = {}
        self.files_to_db["update"] = self.inputs["roc_data_files_to_update"]
        self.files_to_db["insert"] = self.inputs["roc_data_files_to_insert"]

    def run(self):
        logger.debug("LogFileToDb() : run")

        # Get the database connection if needed
        if not hasattr(self, "session"):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Initialize inputs
        self.setup_inputs()

        # Insert / Update files
        for action, files in self.files_to_db.items():
            if files.is_empty:
                continue

            logger.debug("*** TO {} ***".format(action.upper()))
            num_files = len(files.data)
            cpt_file_ok = 0
            cpt_file_ko = 0
            for basename, item in files.data.items():
                logger.debug(item["filepath"])

                try:
                    file_dict = LogFileToDb.item_to_file_log(item, self.root)
                except ValueError as e:
                    logger.error("Value error:", e)
                    if action == "update":
                        # We only store that there was an error
                        file_dict = {"basename": basename, "status": "Failed"}
                    else:
                        cpt_file_ko += 1
                        continue
                except Exception as e:
                    cpt_file_ko += 1
                    logger.error("Unexpected error:", e)
                    continue

                # retrieve parents ID (1/2)
                # remove "parents" item before inserting
                parents = file_dict.pop("parents")

                query = None
                try:
                    if action == "insert":
                        file_log = FileLog(**file_dict)
                        query = self.session.add(file_log)
                    else:
                        # action == 'update'
                        query = (
                            self.session.query(FileLog)
                            .filter(FileLog.basename == basename)
                            .update(file_dict)
                        )
                        file_log = (
                            self.session.query(FileLog)
                            .filter(FileLog.basename == basename)
                            .first()
                        )
                except Exception as e:
                    cpt_file_ko += 1
                    if query:
                        logger.error(
                            "{} file {} has failed: \n {!s}".format(
                                action.capitalize,
                                file_dict["basename"],
                                query.statement.compile(
                                    dialect=postgresql.dialect(),
                                    compile_kwargs={"literal_binds": True},
                                ),
                            )
                        )
                    logger.debug(e)
                    # break to the following file
                    continue
                else:
                    cpt_file_ok += 1
                    logger.info(
                        "{} file {:60s} OK [{:05.2f}% completed]".format(
                            action.capitalize(),
                            file_dict["basename"],
                            100 * cpt_file_ok / num_files,
                        )
                    )

                # ensure that previous versions are set to is_latest=False
                # get file filename without extension and version
                if file_dict["level"] != "SK":
                    trunc_basename = os.path.splitext(file_dict["basename"])[0].replace(
                        "_V" + file_dict["version"], ""
                    )
                    logger.debug(f"TRUNC : {trunc_basename}")

                    base_filter = FileLog.basename.like("%" + trunc_basename + "%")
                    latest_filter = FileLog.is_latest == True  # noqa: E712
                    version_filter = FileLog.version < file_dict["version"]
                    try:
                        query = self.session.query(FileLog).filter(
                            and_(base_filter, latest_filter, version_filter)
                        )

                        for r in query.all():
                            logger.debug(f"OLD VERSION : {r.basename}")

                        old_items = {"is_latest": False}
                        nb = query.update(old_items, synchronize_session=False)
                    except Exception as e:
                        logger.error(f"Error while searching previous versions : {e}")
                    else:
                        if nb > 0:
                            logger.info(f"Setting {nb} files to is_latest = False")

                # retrieve parents ID (2/2)
                parents_added = []
                parents_filter = FileLog.basename.in_(parents)
                parents_file_log = file_log.as_dict()["parents"]
                logger.debug("Parents : {}".format(parents))
                query = None
                try:
                    query = (
                        self.session.query(FileLog)
                        .options(load_only(FileLog.id))
                        .filter(parents_filter)
                    )
                except Exception as e:
                    if query:
                        logger.error(
                            "{} file {} has failed: \n {!s}".format(
                                action.capitalize,
                                file_dict["basename"],
                                query.statement.compile(
                                    dialect=postgresql.dialect(),
                                    compile_kwargs={"literal_binds": True},
                                ),
                            )
                        )
                    logger.debug(e)
                else:
                    for result in query.all():
                        if result.basename not in parents_file_log:
                            file_log.parents.append(result)
                            parents_added.append(result.basename)
                            parents.remove(result.basename)

                            # sometimes a parent without extension is specified
                            # both extensions cdf/h5 are then added
                            # in the parents list
                            # we have to remove the one which wasn't the good one
                            if ".cdf" in result.basename:
                                h5name = result.basename.replace(".cdf", ".h5")
                                if h5name in parents:
                                    parents.remove(h5name)

                            if ".h5" in result.basename:
                                cdfname = result.basename.replace(".h5", ".cdf")
                                if cdfname in parents:
                                    parents.remove(cdfname)

                    # some parents are missing
                    if len(parents) > 0:
                        if file_log.error_log is None:
                            file_log.error_log = ""

                        if file_log.error_log != "":
                            file_log.error_log += "; "

                        file_log.error_log += "Missing parents : " + ", ".join(parents)
                        logger.debug("Missing parents : {}".format(", ".join(parents)))

                    self.session.commit()

                logger.debug("Parents added : {}".format(parents_added))

            logger.info(
                "{} : {} success / {} failure".format(
                    action.capitalize(), cpt_file_ok, cpt_file_ko
                )
            )

    @staticmethod
    def item_to_file_log(item, root):
        """
        Create a dictionary ready to be inserted as a FileLog

        :param item: dictionary with file-system elements already filled :
            size, creation_time, filepath, id
        :param root: the root name of the files (root os not striord in DB)
        :return: dictionary ready to be inserted as a FileLog
        """

        # DATA or SPICE_KERNEL ?
        is_data = True
        url_prefix = ROC_DATA_HTTPS_ROOT
        re_ext = "|".join(DATA_ALLOWED_EXTENSIONS)

        file_name, file_extension = os.path.splitext(item["filepath"])
        if file_extension.replace(".", "") in SPICE_KERNEL_ALLOWED_EXTENSIONS:
            url_prefix = ROC_SPICE_KERNEL_HTTPS_ROOT
            is_data = False
            re_ext = "|".join(SPICE_KERNEL_ALLOWED_EXTENSIONS)

        error_log = []
        warning_log = []

        file_dict = {
            "creation_time": item["creation_time"],
            "size": item["size"],
            "dirname": os.path.dirname(item["filepath"]).replace(root, ""),
            "basename": os.path.basename(item["filepath"]),
            "sha": "",
            "state": "OK",
            "status": "Terminated",
            "insert_time": datetime.now(),
            "descr": "",
            "author": "",
            "is_removed": not os.path.exists(item["filepath"]),
        }

        # Removed file cannot be the latest one
        file_dict["is_latest"] = (not file_dict["is_removed"]) and (
            "former_version" not in file_dict["dirname"]
        )

        file_dict["url"] = (
            url_prefix + "/" + file_dict["dirname"] + "/" + file_dict["basename"]
        )

        # Things in path
        file_path = file_dict["dirname"].split("/")
        if is_data:
            file_dict["level"] = file_path[0]
        else:
            file_dict["level"] = "SK"

        # Things in filename
        if not is_data:
            res = re.search(
                r"^([^\.]+)_V([0-9]+)\.(" + re_ext + ")", file_dict["basename"]
            )
        else:
            res = re.search(
                r"^([^\.]+)_(v|V)([0-9a-zA-Z]+)\.(" + re_ext + ")",
                file_dict["basename"],
            )

        if res:
            file_dict["product"] = res.group(1)
        else:
            file_dict["product"] = ""
            logger.error(
                "Unable to get product name in {}".format(file_dict["basename"])
            )
            error_log.append(
                "Unable to get product name in {}".format(file_dict["basename"])
            )

        file_attrs = {
            "LEVEL": None,
            "TIME_MIN": None,
            "TIME_MAX": None,
            "Dataset_ID": None,
            "Data_version": "00",
            "Parents": [],
            "SPICE_KERNELS": [],
        }

        try:
            if ".h5" in file_dict["basename"]:
                file_attrs = LogFileToDb.get_attrs_from_h5(item["filepath"], file_attrs)

            if ".cdf" in file_dict["basename"]:
                file_attrs = LogFileToDb.get_attrs_from_cdf(
                    item["filepath"], file_attrs
                )

        except Exception as e:
            logger.exception(
                "Error while reading {} : {}".format(file_dict["basename"], e)
            )
            # import traceback
            # traceback.print_exc()
            error_log.append(str(e))

        parents = file_attrs["Parents"]
        start_time = file_attrs["TIME_MIN"]
        end_time = file_attrs["TIME_MAX"]
        data_set_id = file_attrs["Dataset_ID"]
        version = file_attrs["Data_version"]

        # fields error_log and warning_log may have been
        # filled during get_attrs_from_xxx()
        if "error_log" in file_attrs:
            error_log += file_attrs["error_log"]

        if "warning_log" in file_attrs:
            warning_log += file_attrs["warning_log"]

        # Version checks
        if is_data and file_extension != ".xml":
            try:
                # remove leading 'V'
                res = re.search(r"^[v|V]([0-9a-zA-Z]+)$", version)
                if res:
                    logger.debug(f"Removing leading V in version : {version}")
                    version = res.group(1)
            except TypeError:
                logger.error(
                    f"Wrong type for version in {file_dict['basename']}: "
                    f"found {type(version)} but str expected"
                )
                raise

            # check if the one in filename is the one in attributes
            res = re.search(
                r"(v|V)([0-9a-zA-Z]+)\.(" + re_ext + ")", file_dict["basename"]
            )
            if version != "" and res:
                version_in_filename = res.group(2)
                if version != version_in_filename:
                    msg = (
                        f"Versions in filename ({version_in_filename}) and "
                        f"attributes ({version}) do"
                        f" not match in {file_dict['basename']}"
                    )
                    logger.error(msg)
                    error_log.append(msg)
            elif version == "" and res:
                version = res.group(2)
                msg = "No version in attributes"
                logger.warning(msg)
                warning_log.append(msg)
            else:
                version = "00"
                msg = "No version in filename"
                logger.warning(msg)
                warning_log.append(msg)
        else:
            res = re.search(r"_(v|V)([0-9]+)\.(" + re_ext + ")", file_dict["basename"])
            if res:
                version = res.group(2)
            else:
                version = "00"

        logger.debug(f"Version =  {version}")

        # Skipping parents[] search for spice kernels
        if is_data:
            logger.debug("Reading regular data")
            # some parents entries are not well-formed
            to_be_removed = []
            for i, p in enumerate(parents):
                # Some entries are empty
                if re.match(r"\s*$", p):
                    to_be_removed.append(p)
                    continue

                # Some parents have no file extension
                if not re.search(r"\.(\w{,3})$", p):
                    to_be_removed.append(p)
                    # either extensions can be the good one
                    parents.append(p + ".cdf")
                    parents.append(p + ".h5")

                # ignoring mk/*.tm spice kernels (meta kernel)
                if re.search(r"\.tm$", p):
                    to_be_removed.append(p)

            # finally cleaning parents array
            for i, p in enumerate(to_be_removed):
                parents.remove(p)

        # Reading spice kernels
        if not is_data:
            logger.debug("Reading spice kernels")

            start_time = LogFileToDb.commnt_get_meta(item["filepath"], "START_TIME")
            if len(start_time) > 0:
                start_time = min(start_time)
            else:
                start_time = None

            end_time = LogFileToDb.commnt_get_meta(item["filepath"], "STOP_TIME")
            if len(end_time) > 0:
                end_time = min(end_time)
            else:
                end_time = None

        # When start/stop time have not been set
        # Guess them from basename
        datetimes_expr = re.compile(r"_([0-9]{8}T[0-9]{6})\-([0-9]{8}T[0-9]{6})_")
        dates_expr = re.compile(r"_([0-9]{8})\-([0-9]{8})_")
        date_expr = re.compile(r"_([0-9]{8})_")
        if start_time is None or end_time is None:
            res_times = re.search(datetimes_expr, file_dict["basename"])
            res_dates = re.search(dates_expr, file_dict["basename"])
            res_date = re.search(date_expr, file_dict["basename"])

            if res_times:
                try:
                    start_time = datetime.strptime(res_times.group(1), "%Y%m%dT%H%M%S")
                    end_time = datetime.strptime(res_times.group(2), "%Y%m%dT%H%M%S")
                except ValueError as e:
                    logger.warning(
                        f"Unable to parse start time ({res_date.group(1)}) or"
                        f" end time ({res_date.group(2)})  : {e}"
                    )
                    start_time = None
                    end_time = None
            elif res_dates:
                try:
                    start_time = datetime.strptime(res_dates.group(1), "%Y%m%d")
                    end_time = datetime.strptime(res_dates.group(2), "%Y%m%d")
                except ValueError as e:
                    logger.warning(
                        f"Unable to parse start time ({res_date.group(1)}) or"
                        f" end time ({res_date.group(2)})  : {e}"
                    )
                    start_time = None
                    end_time = None

            elif res_date:
                try:
                    start_time = datetime.strptime(res_date.group(1), "%Y%m%d")
                except ValueError as e:
                    logger.warning(
                        f"Unable to parse start time ({res_date.group(1)}) : {e}"
                    )
                    start_time = None
                    end_time = None
                else:
                    end_time = start_time + timedelta(days=1)

            if start_time is not None or end_time is not None:
                logger.info(
                    "Guessing start/end time for {} : {} - {}".format(
                        file_dict["basename"], start_time, end_time
                    )
                )

        if len(error_log) > 0:
            file_dict["state"] = "ERROR"

        elif len(warning_log) > 0:
            file_dict["state"] = "WARNING"

        # Save both error and warning messages
        if (len(error_log) + len(warning_log)) > 0:
            file_dict["error_log"] = "; ".join(error_log + warning_log)
        else:
            # necessary in order to reset error_log when the error is cleared
            file_dict["error_log"] = ""

        # trimming whitespaces
        file_dict["parents"] = [p.strip() for p in parents]

        file_dict["dataset_id"] = data_set_id
        file_dict["version"] = version
        file_dict["start_time"] = start_time
        file_dict["end_time"] = end_time
        file_dict["validity_end"] = None
        file_dict["validity_start"] = None

        # do not set these values here
        # either they are set to false at creation (table default value)
        # or they are set to True with specific tasks
        # file_dict['is_archived'] = False
        # file_dict['is_delivered'] = False

        file_dict["to_update"] = False

        logger.debug(file_dict)

        return file_dict

    @staticmethod
    def commnt_get_meta(filename, key):
        """
        Retrieve parameters in spice kernel using the commnt command

        :param filename: full path of the spice kernel
        :param key: parameter to read
        :return: parameter array values
        """
        val = (
            os.popen(
                "commnt -r " + filename + " | awk -F '=' ' /" + key + "/ {print $2} '"
            )
            .read()
            .split()
        )
        if len(val) == 0:
            if key == "START_TIME":
                val = LogFileToDb.commnt_get_meta(filename, "CK-144000START")
            elif key == "STOP_TIME":
                val = LogFileToDb.commnt_get_meta(filename, "CK-144000STOP")

            val = [datetime.strptime(v, "@%Y-%b-%d-%H:%M:%S.%f") for v in val]

        return val

    @staticmethod
    def get_attrs_from_cdf(filename, file_attrs):
        """
        Retrieve global attributes from CDF file
            :param filename: full path of the file
            :param file_attrs: dictionary with the global attributes to retrieve
            :return: dictionary file_attrs with the attributes filled plus
                error_log and warning_log
        """

        basename = os.path.basename(filename)
        data_set_id = None
        start_time = None
        end_time = None
        version = ""
        parents = []
        parent_version = []
        error_log = []
        warning_log = []

        # Add Parent_version for CDF only
        # NOTE : Parent_versions is an obsolete attribute and
        # should not be found in the new versions of the RPW CDF files.
        # Related instructions are however kept as long as old version CDF files
        # can be processed.
        file_attrs["Parent_version"] = []

        with CDF(filename) as cdf:
            for attr in file_attrs:
                try:
                    logger.debug("{} : {}".format(basename, attr))
                    # don't forget the [...] to get a copy and not a pointer
                    file_attrs[attr] = cdf.attrs[attr][...]
                except KeyError as e:
                    if attr not in OBSOLETE_ATTRS:
                        logger.warning(
                            "Missing attribute in {} : {}".format(
                                basename, re.sub(r"\'", r"", str(e))
                            )
                        )
                        warning_log.append(re.sub(r"\'", r"", str(e)))
                        continue
                else:
                    if attr in OBSOLETE_ATTRS:
                        msg = f"{attr} attribute is obsolete"
                        warning_log.append(msg)
                        logger.warning(msg)

        try:
            start_time = datetime.strptime(
                file_attrs["TIME_MIN"][0], "%Y-%m-%dT%H:%M:%S.%fZ"
            )
        except (ValueError, IndexError, TypeError):
            msg = f"TIME_MIN value is not valid ({file_attrs['TIME_MIN'][0]})"
            logger.error(msg)
            error_log.append(msg)
            logger.debug("Old Julian day format may be used")
        else:
            if start_time < datetime(2020, 2, 10) or start_time > datetime(2100, 1, 1):
                msg = f"End time seems to be erroneous in {basename} : {start_time.isoformat()}"
                logger.error(msg)
                error_log.append(msg)

        try:
            end_time = datetime.strptime(
                file_attrs["TIME_MAX"][0], "%Y-%m-%dT%H:%M:%S.%fZ"
            )
        except (ValueError, IndexError, TypeError):
            msg = f"TIME_MAX value is not valid ({file_attrs['TIME_MAX'][0]})"
            logger.error(msg)
            error_log.append(msg)
            logger.debug("Old Julian day format may be used")
        else:
            if start_time < datetime(2020, 2, 10) or start_time > datetime(2100, 1, 1):
                msg = f"Start time seems to be erroneous in {basename} : {start_time.isoformat()}"
                logger.error(msg)
                error_log.append(msg)

        # Check if data seems to be coherent
        # we sometimes get
        # WARNING: ErfaWarning: ERFA function "d2dtf" yielded
        # 1 of "dubious year (Note 5)" [astropy._erfa.core]

        try:
            data_set_id = file_attrs["Dataset_ID"][0]
            # Clean data_set_id
            # some files have {} around id
            data_set_id = re.sub(r"[\{\}]", "", data_set_id)
        except Exception as e:
            msg = f"Dataset_ID missing in {basename}!"
            logger.error(msg)
            error_log.append(msg)
            logger.debug(e)

        try:
            version = f"{int(file_attrs['Data_version'][0]):02d}"
        except KeyError as e:
            msg = "Data_version attribute missing"
            logger.error(msg)
            error_log.append(msg)
            logger.debug(e)
        except ValueError as e:
            msg = f"Data_version attribute not valid ({file_attrs['Data_version'][0]})"
            logger.error(msg)
            error_log.append(msg)
            logger.debug(e)

        # some files store one parents in each entry,
        # other in string concatenated with commas
        for entry in file_attrs["Parents"]:
            if entry.strip() != "":
                parents += entry.split(",")

        for entry in file_attrs["Parent_version"]:
            # ensure it is a string
            entry = f"{entry}"
            if entry.strip() != "":
                parent_version += entry.split(",")

        # remove extra whitespace
        parents = [p.strip() for p in parents]
        parent_version = [p.strip() for p in parent_version]

        logger.debug("*** PARENTS ***")
        logger.debug(f"{parents}")

        # if parents and parent_version don't have the same length
        # it is useless to try to append the version number
        if 0 < len(parent_version) != len(parents):
            logger.debug(parents)
            logger.debug(parent_version)
            error_log.append("Parents and Parent_version length mismatch")
        elif 0 < len(parent_version) == len(parents):
            # Append the version number if needed
            for i, p in enumerate(parents):
                p_name, p_ext = os.path.splitext(p)
                if not re.search(r"_V[0-9U]+$", p_name):
                    version = parent_version[i]
                    # some files say '5' instead of '05'
                    if len(version) == 1:
                        version = f"0{version}"
                    p_name += f"_V{version}"
                    parents[i] = p_name + p_ext
                    logger.debug(f"Adding version to parent name : {parents[i]}")

        # some files store spice kernels in array,
        # other in string concatenated with commas
        for entry in file_attrs["SPICE_KERNELS"]:
            parents += entry.split(",")

        for i, p in enumerate(parents):
            if p.find("CDF>") != -1:
                parents[i] = p.replace("CDF>", "")
            if p.find("L0>") != -1:
                parents[i] = p.replace("L0>", "")
            if p.find("ANC>") != -1:
                parents[i] = p.replace("ANC>", "")

        file_attrs["TIME_MIN"] = start_time
        file_attrs["TIME_MAX"] = end_time
        file_attrs["Dataset_ID"] = data_set_id
        file_attrs["Parents"] = parents
        file_attrs["Data_version"] = version

        file_attrs["error_log"] = error_log
        file_attrs["warning_log"] = warning_log

        return file_attrs

    @staticmethod
    def get_attrs_from_h5(filename, file_attrs):
        """
        Retrieve attrs from HDF5 file
            :param filename: full path of the file
            :param file_attrs: dictionary with the attributes to retrieve
            :return: dictionary file_attrs with the attributes filled plus
                error_log and warning_log
        """

        basename = os.path.basename(filename)
        version = ""
        error_log = []
        warning_log = []

        with h5py.File(filename, "r") as l0:
            for attr in file_attrs:
                try:
                    logger.debug("{} : {}".format(basename, attr))
                    file_attrs[attr] = l0.attrs[attr]
                except KeyError as e:
                    if attr not in OBSOLETE_ATTRS:
                        logger.warning(
                            "Missing attribute in {} : {}".format(basename, e)
                        )
                        warning_log.append(str(e))
                        continue
                else:
                    if attr in OBSOLETE_ATTRS:
                        msg = f"{attr} is obsolete"
                        logger.warning(msg)
                        warning_log.append(msg)

        # Get TIME_MIN/TIME_MAX L0 attributes value as datetime
        file_attrs["TIME_MIN"] = datetime.strptime(
            file_attrs["TIME_MIN"], TIME_ISO_STRFORMAT
        )

        file_attrs["TIME_MAX"] = datetime.strptime(
            file_attrs["TIME_MAX"], TIME_ISO_STRFORMAT
        )

        file_attrs["Parents"] = file_attrs["Parents"].split(",")
        if len(file_attrs["SPICE_KERNELS"]) > 0:
            file_attrs["Parents"] += file_attrs["SPICE_KERNELS"].split(",")

        # Ensure data version is a string on 2 digits
        try:
            version = f"{file_attrs['Data_version']:02d}"
        except KeyError:
            msg = "Data_version missing"
            logger.error(msg)
            error_log.append(msg)
        except ValueError:
            msg = f"Data_version is not valid ({file_attrs['Data_version']})"
            logger.error(msg)
            error_log.append(msg)

        file_attrs["Data_version"] = version
        file_attrs["error_log"] = error_log
        file_attrs["warning_log"] = warning_log

        return file_attrs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import os


from sqlalchemy import and_
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import load_only

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import PIPELINE_DATABASE, ROC_DATA_ROOT
from roc.dingo.models.file import FileLog


__all__ = ["SearchMissingParentsFiles"]


class SearchMissingParentsFiles(Task):
    """
    Parse ROC db and search for files with Missing Parents in the error_log
    """

    plugin_name = "roc.dingo"
    name = "search_missing_parents_files"
    files_to_update = []

    def add_targets(self):
        logger.debug("SearchMissingParentsFiles() : add_targets")

        self.add_output(
            target_class=BaseTarget, many=True, identifier="roc_data_files_to_update"
        )

        # useless here but necessary for LogFileToDb()
        self.add_output(
            target_class=BaseTarget, many=True, identifier="roc_data_files_to_insert"
        )

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug("SearchMissingParentsFiles() : setup_inputs")

        # get the root file tree
        self.root = self.pipeline.get("root", default=ROC_DATA_ROOT, args=True)
        # ensure that there is / at the end
        self.root = os.path.join(self.root, "")

    def run(self):
        logger.debug("SearchMissingParentsFiles() : run")

        # Get the database connection if needed
        if not hasattr(self, "session"):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Initialize inputs
        self.setup_inputs()

        # Initialize list of ROC data files to update into db
        file_list = {}

        logger.debug("Root directory: {}".format(self.root))

        # Define relevant fields to retrieve
        fields = [
            getattr(FileLog, f)
            for f in ["basename", "dirname", "creation_time", "size"]
        ]

        # Search files with "Missing parents in error_log"
        missing_filter = FileLog.error_log.like("%Missing parents%")
        not_removed_filter = FileLog.is_removed == False  # noqa: E712
        query = None
        try:
            query = (
                self.session.query(FileLog)
                .options(load_only(*fields))
                .filter(and_(missing_filter, not_removed_filter))
            )
            logger.debug(
                str(
                    query.statement.compile(
                        dialect=postgresql.dialect(),
                        compile_kwargs={"literal_binds": True},
                    )
                )
            )
            results = query.all()
        except Exception as e:
            if query:
                logger.error(
                    "Query has failed: \n {!s}".format(
                        query.statement.compile(
                            dialect=postgresql.dialect(),
                            compile_kwargs={"literal_binds": True},
                        )
                    )
                )
            logger.debug(e)
        else:
            for item in results:
                # adding an entry in file_list
                file_list[item.basename] = {
                    "filepath": os.path.join(self.root, item.dirname, item.basename),
                    "size": item.size,
                    "creation_time": item.creation_time,
                }

        # Returns outputs
        self.outputs["roc_data_files_to_update"].data = file_list
        logger.info("{} files to update".format(str(len(file_list))))
        if len(file_list) == 0:
            self.outputs["roc_data_files_to_update"].is_empty = True

        self.outputs["roc_data_files_to_insert"].data = {}
        self.outputs["roc_data_files_to_insert"].is_empty = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to insert SOLO HK EDDS data into the ROC database."""

from datetime import datetime, timedelta
import hashlib

import pandas as pd
from sqlalchemy import and_
import xmltodict

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import FileTarget
from sqlalchemy.exc import IntegrityError

from roc.dingo.constants import (
    PIPELINE_DATABASE,
    TIME_SQL_STRFORMAT,
    TRYOUTS,
    TIME_WAIT_SEC,
    SQL_LIMIT,
)
from roc.dingo.models.data import SoloHkParam, ProcessQueue
from roc.dingo.tools import (
    query_db,
    get_columns,
    get_or_create_in_db,
    bulk_insert,
    glob_paths,
)

__all__ = ["SoloHkToDb"]


class SoloHkToDb(Task):
    """
    Insert the content of input SOLO HK EDDS data files into the ROC database
    """

    plugin_name = "roc.dingo"
    name = "solohk_to_db"

    def add_targets(self):
        self.add_input(
            identifier="solo_hk_files",
            many=True,
            filepath=SoloHkToDb.get_solohk_files,
            target_class=FileTarget,
        )

    @staticmethod
    def get_solohk_files(pipeline):
        try:
            solo_hk_files = pipeline.args.solo_hk_files
            if not isinstance(solo_hk_files, list):
                solo_hk_files = [solo_hk_files]
            return solo_hk_files
        except Exception as e:
            # If not defined as input argument, then assume that it is already
            # defined as target input
            logger.debug(e)
            pass

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):
        # get the input SOLO HK files
        self.solo_hk_files = glob_paths(self.inputs["solo_hk_files"].filepath)

        # Get or create failed_files list from pipeline properties
        self.failed_files = self.pipeline.get("failed_files", default=[], create=True)

        # Get or create processed_files list from pipeline properties
        self.processed_files = self.pipeline.get(
            "processed_files", default=[], create=True
        )

        # Get tryouts from pipeline properties
        self.tryouts = self.pipeline.get("tryouts", default=[TRYOUTS], create=True)[0]

        # Get wait from pipeline properties
        self.wait = self.pipeline.get("wait", default=[TIME_WAIT_SEC], create=True)[0]

        # Retrieve --limit keyword value
        self.limit = self.pipeline.get("limit", default=[SQL_LIMIT], args=True)[0]

        # Get --to-queue keyword
        self.to_queue = self.pipeline.get("to_queue", default=False, create=True)

        # get a database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get columns of the table
        self.model = SoloHkParam
        self.columns = get_columns(self.model, remove=["id"])

        # Get today date time
        self.today = datetime.today()

        # Initialize task counters
        self.inserted_counter = 0
        self.invalid_counter = 0

    def run(self):
        logger.debug("[SoloHkToDb]: Task is starting")
        try:
            self.setup_inputs()
        except Exception as e:
            logger.error("[SoloHkToDb]: Initializing inputs has failed!")
            logger.debug(e)
            self.pipeline.exit()
            return

        n_files = len(self.solo_hk_files)
        logger.info(f"{n_files} input solo_hk_files to process")
        if n_files == 0:
            return

        # Loop over each input file in the input list
        for i, current_file in enumerate(self.solo_hk_files):
            logger.info(f"Processing {current_file}  ({n_files - i - 1} remaining)")

            # Parse input file
            try:
                current_data = self.parse_solohk_xml(current_file)
            except Exception as e:
                logger.error(f"Parsing {current_file} has failed!")
                logger.debug(e)
                self.failed_files.append(current_file)
                continue

            n_element = current_data.shape[0]
            logger.info(
                f"{n_element} <ParamSampleListElement> elements found in {current_file}"
            )
            if n_element == 0:
                logger.info(f"{current_file} is empty, skip it")
                self.processed_files.append(current_file)
                continue

            # Get list of pre-existing elements in the database
            try:
                # First convert input XML fields into
                # expected table column names
                # and compute extra values (SHA, utc_time)
                current_data = self.xml_to_solohkparam(current_data)
            except Exception as e:
                logger.error(f"Preprocessing data from {current_file} has failed!")
                logger.debug(e)
                self.failed_files.append(current_file)
                continue
            else:
                n_data_to_insert = current_data.shape[0]

            logger.info(f"Inserting {n_data_to_insert} elements from {current_file}")
            if n_data_to_insert == 0:
                self.processed_files.append(current_file)
                continue
            else:
                data_to_insert = current_data[self.columns].to_dict("records")

            try:
                bulk_insert(self.session, self.model, data_to_insert)
            except IntegrityError:
                logger.debug(
                    f"Some data already found in the database for "
                    f"{current_file}, attempting to insert new data only"
                )
                # Entries already found in the database, try to insert only new data
                # First query existing data (only SHA is required)
                # Define time range of request
                start_time = current_data["utc_time"].min()
                end_time = current_data["utc_time"].max()
                existing_data = pd.DataFrame.from_records(
                    self._get_existing_data(
                        self.model,
                        self.model.sha,
                        start_time=start_time,
                        end_time=end_time,
                    )
                )

                # Only keep unique SHA elements
                current_data = current_data[~current_data.sha.isin(existing_data.sha)]

                n_data_to_insert = current_data.shape[0]
                if n_data_to_insert == 0:
                    logger.info(f"No new data to insert for {current_file}")
                    self.processed_files.append(current_file)
                    continue
                else:
                    # Re-insert only new elements
                    data_to_insert = current_data[self.columns].to_dict("records")
                    bulk_insert(self.session, self.model, data_to_insert)
                    logger.debug(
                        f"{n_data_to_insert} new elements inserted for {current_file}"
                    )
            except Exception as e:
                logger.error(f"Inserting new data from {current_file} has failed!")
                logger.debug(e)
                self.failed_files.append(current_file)
                continue
            else:
                self.processed_files.append(current_file)
                # Increment number of elements inserted
                self.inserted_counter += n_data_to_insert

            # Add current element days to process_queue table
            if self.to_queue:
                # Only add to process_queue table
                # new days to insert
                days_to_process = list(
                    set(
                        [
                            datetime.combine(current_date, datetime.min.time())
                            for current_date in current_data["utc_time"].to_list()
                        ]
                    )
                )

                for current_day in days_to_process:
                    current_entry = self._as_process_queue(current_day)
                    job, done, created = get_or_create_in_db(
                        self.session, ProcessQueue, current_entry
                    )
                    if done:
                        logger.debug(f"{current_entry} inserted in process_queue table")
                    else:
                        logger.error(
                            f"{current_entry} cannot be inserted in process_queue table"
                        )

        n_processed = len(self.processed_files)
        n_failed = len(self.failed_files)
        if n_processed > 0:
            logger.info(
                f"{self.inserted_counter} new elements inserted from {n_processed} files"
            )
        if n_failed > 0:
            logger.error(f"Insertion has failed for {n_failed} files!")

    def _get_existing_data(self, model, fields, start_time=None, end_time=None):
        """
        Query database to return existing data for a given table

        :param model: class of the table
        :param fields: fields to query
        :return: returned rows (as a list of dictionaries)
        """

        # Get list of existing data in the database
        filters = []
        # Add start_time/end_time filters (if passed)
        if start_time:
            filters.append(model.utc_time >= str(start_time - timedelta(hours=1)))
        if end_time:
            filters.append(model.utc_time <= str(end_time + timedelta(hours=1)))

        if fields is None:
            fields = model

        results = query_db(
            self.session,
            fields,
            filters=and_(*filters),
            tryouts=self.tryouts,
            wait=self.wait,
            limit=self.limit,
            to_dict="records",
        )
        return results

    def parse_solohk_xml(self, solo_hk_xml):
        """
        Parse input SOLO HK EDDS XML file.

        :param solo_hk_xml: Path of the input Solo HK EDDS file
        :return: List of <ParamSampleListElement> elements
        """
        logger.debug(f"Parsing {solo_hk_xml} ...")
        with open(solo_hk_xml, "r") as xml:
            xml_data = xmltodict.parse(xml.read())["ns2:ResponsePart"]["Response"][
                "ParamResponse"
            ]["ParamSampleList"]["ParamSampleListElement"]

        if not isinstance(xml_data, list):
            xml_data = [xml_data]

        xml_data = pd.DataFrame.from_records(xml_data)

        return xml_data

    def xml_to_solohkparam(self, xml_data):
        """
        Convert <ParamSampleListElement> XML elements
        as entries to be inserted in the solo_hk_param table of the ROC database

        :param xml_data: pandas.Dataframe containing the <ParamSampleListElement> elements from input XML file
        :return: pandas.Dataframe with expected fields values for solo_hk_param table
        """
        # First keep only valid elements
        solohk_data = xml_data.loc[xml_data["Validity"].isin(["VALID"])]
        solohk_data.reset_index(drop=True, inplace=True)

        if not solohk_data.shape[0] == 0:
            # Define field names as expected for solo_hk_param table
            solohk_data = solohk_data.rename(
                columns={
                    "Name": "name",
                    "Unit": "unit",
                    "Description": "description",
                    "EngineeringValue": "eng_value",
                    "RawValue": "raw_value",
                },
                inplace=False,
            )

            # Convert elements times in datetime object
            solohk_data["utc_time"] = solohk_data.apply(
                lambda x: datetime.strptime(x["TimeStampAsciiA"], TIME_SQL_STRFORMAT),
                axis=1,
            )
            # Compute SHA256 of the elements
            solohk_data["sha"] = solohk_data.apply(self._get_sha, axis=1)

            # Make sure to have unique elements (unique SHA values)
            solohk_data.drop_duplicates(subset=["sha"], inplace=True)

        return solohk_data

    def _get_sha(self, data):
        """
        Compute the SHA of the input element (from all parameters in data)

        :param data: Current SOLO HK parameter element
        :return: associated SHA256
        """
        sha = hashlib.sha256()
        sha.update(data["utc_time"].isoformat().encode("utf-8"))
        sha.update(data["name"].encode("utf-8"))
        sha.update(data["description"].encode("utf-8"))
        sha.update(data["raw_value"].encode("utf-8"))
        sha.update(data["eng_value"].encode("utf-8"))
        return str(sha.hexdigest())

    def _as_process_queue(self, current_date):
        """
        Return expected format for a process_queue table entry
        from input current date

        :param current_date: datetime.date() object to convert
        :return: dictionary for data_queue table entry
        """
        start_time = datetime.combine(current_date, datetime.min.time())
        return {
            "dataset_id": "SOLO_HK_PLATFORM",
            "start_time": start_time,
            "insert_time": self.today,
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""

import pytest
import os.path as osp
import yaml
from argparse import Namespace

from poppy.core.pipeline import Pipeline
from poppy.core.logger import logger
from poppy.core.test import TaskTestCase

from roc.dingo.tasks import LogFileToDb
from roc.dingo.models.file import FileLog
from roc.dingo.tests.test_log_file import DingoLogFileTest


# Tests on roc.dingo.tasks.log_file methods
class TestDingoLogFileToDb(TaskTestCase):
    @pytest.mark.parametrize(
        "root, data_file, expected_data_dump",
        [
            (
                "logfile_to_rocdb/DATA-1/",
                "roc_data_files_all.dat",
                "roc_data_files_all_dump.yml",
            ),
        ],
    )
    def test_log_file_to_db(self, root, data_file, expected_data_dump):
        """
        Test the DINGO LogFileToDb task with standard parameters
        Start from an empty database
        Give a roc_data_files list
        All items should be inserted

        :return:
        """

        logger.info("test_log_file_to_db() ")

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, dry_run=False)

        # --- database setup ---
        session = DingoLogFileTest.setup_session()

        # --- define data test --
        data1 = DingoLogFileTest.open_test_data(data_file, root)

        # --- initialise task & pipeline --
        self.task = LogFileToDb()
        self.task.pipeline = Pipeline(args)
        self.task.instantiate_targets()
        self.task.session = session

        # --- initialise inputs ---
        self.task.inputs["roc_data_files_to_insert"].data = data1.copy()

        self.task.inputs["roc_data_files_to_update"].data = {}
        self.task.inputs["roc_data_files_to_update"].is_empty = True

        # --- run the task ---
        self.run_task()

        # --- get results ---
        results = session.query(FileLog).all()

        # --- read expected data
        path = osp.abspath(osp.dirname(__file__))
        path = osp.join(path, "../data-tasks", expected_data_dump)
        with open(path) as f:
            expected_data = yaml.load(f, Loader=yaml.FullLoader)

        # --- make assertions ---
        # all files should have been inserted
        assert len(results) == len(expected_data)

        for item in results:
            assert item.basename in expected_data
            assert item.as_dict() == expected_data[item.basename]

        # --- close session
        session.close()

    @pytest.mark.parametrize(
        "root, data_file, expected_data_dump",
        [
            ("logfile_to_rocdb/SK-1/", "roc_sk_files_1.dat", "roc_sk_files_1_dump.yml"),
        ],
    )
    def test_log_file_sk_to_db(self, root, data_file, expected_data_dump):
        """
        Test the DINGO LogFileToDb task with standard parameters
        Start from an empty database
        Give a roc_data_files spice kernel list
        All items should be inserted

        :return:
        """

        logger.info("test_log_file_sk_to_db() ")

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, dry_run=False)

        # --- database setup ---
        session = DingoLogFileTest.setup_session()

        # --- define data test --
        data1 = DingoLogFileTest.open_test_data(data_file, root)

        # --- initialise task & pipeline --
        self.task = LogFileToDb()
        self.task.pipeline = Pipeline(args)
        self.task.instantiate_targets()
        self.task.session = session

        # --- initialise inputs ---
        self.task.inputs["roc_data_files_to_insert"].data = data1.copy()

        self.task.inputs["roc_data_files_to_update"].data = {}
        self.task.inputs["roc_data_files_to_update"].is_empty = True

        # --- run the task ---
        self.run_task()

        # --- get results ---
        results = session.query(FileLog).filter(FileLog.level == "SK").all()

        # --- read expected data
        path = osp.abspath(osp.dirname(__file__))
        path = osp.join(path, "../data-tasks", expected_data_dump)
        with open(path) as f:
            expected_data = yaml.load(f, Loader=yaml.FullLoader)

        # --- make assertions ---
        # all files should have been inserted
        assert len(results) == len(expected_data)

        for item in results:
            assert item.basename in expected_data
            assert item.as_dict() == expected_data[item.basename]

        # --- close session
        session.close()

    @pytest.mark.parametrize(
        "root, data_file, file, old_file",
        [
            (
                "logfile_to_rocdb/DATA-1/",
                "roc_data_files_2020_08_19.dat",
                "solo_L1_rpw-lfr-surv-bp1-cdag_20200819_V07.cdf",
                "solo_L1_rpw-lfr-surv-bp1-cdag_20200819_V06.cdf",
            ),
        ],
    )
    def test_add_file_with_old_version_removed(self, root, data_file, file, old_file):
        """
        https://gitlab.obspm.fr/ROC/Pipelines/Plugins/DINGO/-/issues/23
        Some files are removed but still indicated as latest

        Test the DINGO CheckFileInDb task with standard parameters
        Start from an empty database
        Populates it by hand with roc_data_files data
        Remove a file from the list
        Start the task
        Removed file is flagged as removed but still latest
        Give a new data file containing a new version of the file removed
        The removed version should be flagged as not latest

        :return:
        """

        logger.info("test_add_file_with_old_version_removed() ")

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, dry_run=False)

        # --- database setup ---
        session = DingoLogFileTest.setup_session()

        # --- define data test --
        data1 = DingoLogFileTest.open_test_data(data_file, root)

        # replace the old file by the new file
        item = data1.pop(file)
        item["filepath"] = item["filepath"].replace(file, old_file)
        data1[old_file] = item

        # populate database
        for item, data in data1.items():
            file_log = DingoLogFileTest.data_to_file_log(data, root)
            # mark the old file as removed
            if file_log.basename == old_file:
                file_log.is_removed = True
            session.add(file_log)
        session.flush()

        # new file has to be inserted
        item = data1.pop(old_file)
        item["filepath"] = item["filepath"].replace(old_file, file)
        data2 = {}
        data2[file] = item

        # --- initialise task & pipeline --
        self.task = LogFileToDb()
        self.task.pipeline = Pipeline(args)
        self.task.instantiate_targets()
        self.task.session = session

        # --- initialise inputs ---
        self.task.inputs["roc_data_files_to_insert"].data = data2.copy()

        self.task.inputs["roc_data_files_to_update"].data = {}
        self.task.inputs["roc_data_files_to_update"].is_empty = True

        # --- run the task ---
        self.run_task()

        results = session.query(FileLog).filter(FileLog.basename == file)
        for item in results:
            assert not item.is_removed
            assert item.is_latest

        results = session.query(FileLog).filter(FileLog.basename == old_file)
        for item in results:
            assert item.is_removed
            assert not item.is_latest

        # --- close session
        session.close()



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0022_add_to_update_flag_to_filelog
Revises: roc_dingo_0021_add_retrieved_time
Create Date: 2022-01-11 16:07:33.943889

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "roc_dingo_0022_add_to_update_flag_to_filelog"
down_revision = "roc_dingo_0021_add_retrieved_time"
branch_labels = None
depends_on = None


def upgrade():
    # Add "to_update" flag with a default value to false
    # Trick to set a default value to the non null column
    # https://gist.github.com/VladimirPal/e256d19df8d85234dde1
    # it avoids to create with nullable=true, set the colum to false and
    # then set nullable=false
    op.add_column(
        "file_log",
        sa.Column(
            "to_update",
            sa.BOOLEAN(),
            server_default=sa.schema.DefaultClause("0"),
            nullable=False,
        ),
        schema="pipeline",
    )

    # Ensure is_removed and is_delivered are set to False

    op.alter_column(
        "file_log",
        "is_archived",
        server_default=sa.schema.DefaultClause("0"),
        schema="pipeline",
    )

    op.alter_column(
        "file_log",
        "is_delivered",
        server_default=sa.schema.DefaultClause("0"),
        schema="pipeline",
    )

    # ### end Alembic commands ###


def downgrade():
    op.drop_column("file_log", "to_update", schema="pipeline")

    op.alter_column("file_log", "is_archived", server_default=None, schema="pipeline")

    op.alter_column("file_log", "is_delivered", server_default=None, schema="pipeline")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to query the pipeline database."""

import json
import os

from sqlalchemy import and_

from roc.dingo.tools import query_db, valid_time

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.conf import settings

from roc.dingo.constants import (
    PIPELINE_DATABASE,
    SQL_LIMIT,
    PIPELINE_TABLES,
    TIME_OUTPUT_STRFORMAT,
)

__all__ = ["ExportToJson"]

# ________________ Global Variables _____________
# (define here the global variables)


# ________________ Class Definition __________
# (If required, define here classes)


class ExportToJson(Task):
    """
    Task to export content of the database into a JSON file.
    """

    plugin_name = "roc.dingo"
    name = "export_to_json"

    @Connector.if_connected(PIPELINE_DATABASE)
    def run(self):
        # Initialize output dictionary
        output_json_dict = {}

        # get the pipeline connector and get the database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get list of pipeline database tables to query
        table_list = self.pipeline.get(
            "tables", default=list(PIPELINE_TABLES.keys()), args=True
        )

        # If --start_time keyword value is set, add an lower time limit
        start_time = valid_time(
            self.pipeline.get("start_time", default=[None], args=True)[0]
        )

        # If --end_time keyword value is set, add an upper time limit
        end_time = valid_time(
            self.pipeline.get("end_time", default=[None], args=True)[0]
        )

        # If --creation-time is set to True, then
        # filtering packet_log and file_log table entries using creation time
        # instead of insertion time (not possible for invalid_packet_log)
        creation_time_flag = self.pipeline.get(
            "creation-time", default=False, args=True
        )

        for current_table in table_list:
            logger.info(f"Exporting entries for {current_table}")

            # Get table model
            current_model = PIPELINE_TABLES[current_table]

            if current_table == "packet_log":
                # Define time to use for start_time/end_time filter
                # Default is insertion time
                if creation_time_flag:
                    time = current_model.utc_time
                else:
                    time = current_model.insertion_time
            elif current_table == "invalid_packet_log":
                time = current_model.insertion_time
            elif current_table == "file_log":
                if creation_time_flag:
                    time = current_model.file_creation_date
                else:
                    time = current_model.file_insert_date
            else:
                logger.warning(f"Unknown table: {current_table}")
                continue

            # Add input filters
            filters = []

            if start_time:
                filters.append(time >= start_time)  # greater or equal than ...

            if end_time:
                filters.append(time < end_time)  # lesser than ...

            filters = and_(*filters)

            # Retrieve --limit keyword value
            limit = self.pipeline.get("limit", default=[SQL_LIMIT], args=True)[0]

            # get entries from pipeline table in ROC database
            # Order by increasing time
            entries_dict = query_db(
                self.session,
                current_model,
                filters=filters,
                limit=limit,
                order_by=time,
                to_dict="records",
            )
            entry_count = len(entries_dict)
            if entry_count == limit:
                logger.warning(f"Row limit has been reached {limit}")
            elif entry_count == 0:
                logger.info(f"No entry found in {current_model}")
            else:
                output_json_dict[current_table] = entries_dict

        if not output_json_dict:
            return

        # Build output file path
        output_json = self.pipeline.get("output_json", default=None)
        if output_json is None:
            database_name = "rocdb"
            for database in self.pipeline.properties.configuration[
                "pipeline.databases"
            ]:
                if database["identifier"] == settings.PIPELINE_DATABASE:
                    database_name = database["login_info"]["database"]
                    break

            filename_field_list = [database_name, "-".join(table_list)]
            if start_time:
                filename_field_list.append(start_time.strftime(TIME_OUTPUT_STRFORMAT))
            if end_time:
                filename_field_list.append(end_time.strftime(TIME_OUTPUT_STRFORMAT))
            output_json = "_".join(filename_field_list) + ".json"

        output_json_path = os.path.join(self.pipeline.output, output_json)
        # Write output JSON file
        with open(output_json_path, "w") as json_buff:
            json.dump(output_json_dict, json_buff)
        logger.info(f"{output_json_path} saved")


# ________________ Global Functions __________
# (If required, define here global functions)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import re

from sqlalchemy import text

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task

from roc.dingo.constants import PIPELINE_DATABASE
from roc.dingo.models.file import FileLog


__all__ = ["UpdateParentsInDb"]


class UpdateParentsInDb(Task):
    """
    Update database linking files and their parents, based on the
    error_log field containing "Missing parents"
    """

    plugin_name = "roc.dingo"
    name = "update_parents_in_db"

    def add_targets(self):
        pass

    def setup_inputs(self):
        pass

    @staticmethod
    def remove_parent_from_log(child, parent):
        """
        Take a pair child/parent and remove the child error_log
        part containing the parent message

        :param child: FileLog element for child
        :param parent: FileLog element for parent
        """

        miss_expr = re.compile("Missing parents : (.*)$")

        logger.debug("Child basename : {}".format(child.basename))
        logger.debug("Child error_log : {}".format(child.error_log))
        logger.debug("Parent basename : {}".format(parent.basename))

        # search into the error_log message
        old_missing_parents = miss_expr.search(child.error_log)
        missing_parents = old_missing_parents.group(1).split(", ")

        logger.debug("Child missing parents : {}".format(missing_parents))

        # remove parent error message
        if parent.basename in missing_parents:
            missing_parents.remove(parent.basename)
        else:
            logger.warning(
                "Parent {} not found in error_log '{}'. This should not happen.".format(
                    parent.basename, child.error_log
                )
            )

        # if the parent has .h5 or .cdf as extension
        # try to remove the other extension (both are in the error_log)

        if ".cdf" in parent.basename:
            h5name = parent.basename.replace(".cdf", ".h5")
            if h5name in missing_parents:
                missing_parents.remove(h5name)

        if ".h5" in parent.basename:
            cdfname = parent.basename.replace(".h5", ".cdf")
            if cdfname in missing_parents:
                missing_parents.remove(cdfname)

        # restore the error_log message
        if len(missing_parents) > 0:
            missing_parents = "Missing parents : " + ", ".join(missing_parents)
        else:
            missing_parents = ""

        # update error_log in DB
        child.error_log = child.error_log.replace(
            old_missing_parents.group(0), missing_parents
        )

    def run(self):
        logger.debug("UpdateParentsInDb() : run")

        # Get the database connection if needed
        if not hasattr(self, "session"):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Update Database with
        # the pairs (child/parent) which can be linked
        # based on the error_log
        # Returns the pairs inserted

        sql1 = text("""
            with f1 (id, file, parent) as (select distinct f.id, f.basename ,
                trim(unnest(string_to_array(substring(substring(f.error_log
                    from 'Missing parents : .*$') from 19), ', ')))
                from pipeline.file_log f
                where f.is_removed = False and f.error_log ~ 'Missing parents')
            insert into pipeline.parents_file_log (id_child, id_parent) (
                select f1.id , f2.id
                from f1 inner join pipeline.file_log f2
                on f1.parent = f2.basename
                except (
                    select id_child, id_parent from pipeline.parents_file_log))
            -- on conflict do nothing
            -- to be restored when bdd-lesia will be > 9.4
            returning id_child, id_parent;
            """)

        results1 = self.session.execute(sql1)

        cpt = 0
        for row in results1:
            cpt += 1
            file = self.session.get(FileLog, row[0])
            parent = self.session.get(FileLog, row[1])

            logger.debug(row)
            logger.info(
                "Add parent {:60s} to {:60s} [{:05.2f}% complete]".format(
                    parent.basename, file.basename, 100 * cpt / results1.rowcount
                )
            )

            UpdateParentsInDb.remove_parent_from_log(file, parent)

        # Returns the pairs that could be linked based on the error_log
        # but are already linked in parents_file_log
        # This should not happen, but may be useful after a pipeline error

        sql2 = text("""
            with f1 (id, file, parent) as (select distinct f.id, f.basename ,
                trim(unnest(string_to_array(substring(substring(f.error_log
                    from 'Missing parents : .*$') from 19), ', ')))
                from pipeline.file_log f
                where f.is_removed = False and f.error_log ~ 'Missing parents')
            select f1.id , f2.id
                from f1 inner join pipeline.file_log f2
                on f1.parent = f2.basename
                intersect (
                    select id_child, id_parent from pipeline.parents_file_log)
            """)

        # ensure that preceding instructions have been committed
        self.session.commit()

        results2 = self.session.execute(sql2)

        cpt = 0
        for row in results2:
            cpt += 1
            file = self.session.get(FileLog, row[0])
            parent = self.session.get(FileLog, row[1])

            logger.debug(row)
            logger.info(
                "Parent {:60s} to {:60s} was already added [{:05.2f}% complete]".format(
                    parent.basename, file.basename, 100 * cpt / results2.rowcount
                )
            )

            UpdateParentsInDb.remove_parent_from_log(file, parent)

        if results1.rowcount + results2.rowcount == 0:
            logger.info("Nothing to be done.")



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0008_remove_date
Revises: roc_dingo_0007_add_error_log
Create Date: 2021-02-04 16:17:41.195180

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "roc_dingo_0008_remove_date"
down_revision = "roc_dingo_0007_add_error_log"
branch_labels = None
depends_on = None


def upgrade():
    # Column date is useless.
    # All the information is in start/end time`
    # date column is irrelevant for monthly files.
    op.drop_column("file_log", "date", schema="pipeline")


def downgrade():
    #
    op.add_column(
        "file_log", sa.Column("date", sa.DATE(), nullable=True), schema="pipeline"
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to export SOLO HK EDDS data from the ROC database."""

import os
from datetime import timedelta, datetime
from pathlib import Path

from sqlalchemy import asc

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import FileTarget

from roc.dingo.constants import (
    PIPELINE_DATABASE,
    START_TIME,
    END_TIME,
    TRYOUTS,
    TIME_WAIT_SEC,
    SQL_LIMIT,
    DATA_VERSION,
    JENV,
    SOLO_HK_TEMPLATE,
    TIME_DAILY_STRFORMAT,
    TIME_SQL_STRFORMAT,
)
from roc.dingo.models.data import ProcessQueue, SoloHkParam
from roc.dingo.tools import valid_data_version, query_db, delete_in_db, valid_time

__all__ = ["DbToSoloHk"]


class DbToSoloHk(Task):
    """
    Export solo_hk_param table data of the ROC database into output daily files.
    The output file format is a light version of SOLO EDDS response XML.
    """

    plugin_name = "roc.dingo"
    name = "db_to_solohk"

    def add_targets(self):
        self.add_output(identifier="solo_hk_files", many=True, target_class=FileTarget)

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):
        # Get start_time input value
        self.start_time = valid_time(
            self.pipeline.get("start_time", default=[None], create=True)[0]
        )
        if not self.start_time:
            self.start_time = START_TIME

        # Get end_time input value
        self.end_time = valid_time(
            self.pipeline.get("end_time", default=[None], create=True)[0]
        )
        if not self.end_time:
            self.end_time = END_TIME

        # Get list of dates to process
        # (If passed with start_time/end_time keywords, then only
        # process dates between start_time and end_time)
        self.date_list = self.pipeline.get("date_list", default=[])

        # Get tryouts from pipeline properties
        self.tryouts = self.pipeline.get("tryouts", default=[TRYOUTS], create=True)[0]

        # Get wait from pipeline properties
        self.wait = self.pipeline.get("wait", default=[TIME_WAIT_SEC], create=True)[0]

        # Get or create failed_files list from pipeline properties
        self.failed_files = self.pipeline.get("failed_files", default=[], create=True)

        # Retrieve --limit keyword value
        self.limit = self.pipeline.get("limit", default=[SQL_LIMIT], args=True)[0]

        # Retrieve --limit keyword value
        self.data_version = valid_data_version(
            self.pipeline.get("data_version", default=[DATA_VERSION], args=True)[0]
        )

        # Get output directory
        self.output_dir = self.pipeline.output
        if not self.output_dir:
            raise FileNotFoundError(
                f"[DbToSoloHk]: output directory is not defined or does not exist ({self.output_dir})"
            )
        if not os.path.isdir(self.output_dir):
            Path(self.output_dir).mkdir(parents=True, exist_ok=True)

        # Get overwrite optional keyword
        self.overwrite = self.pipeline.get("overwrite", default=False, args=True)

        # Get from-queue optional keyword
        self.from_queue = self.pipeline.get("from_queue", default=False)

        # Get clean-queue optional keyword
        self.clean_queue = self.pipeline.get("clean_queue", default=False)

        # get a database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

    def run(self):
        logger.debug("[DbToSoloHk]: Task is starting")
        try:
            self.setup_inputs()
        except Exception as e:
            logger.error("[DbToSoloHk]: Initializing inputs has failed!")
            logger.debug(e)
            self.pipeline.exit()
            return

        # if --from-queue is set get list of dates to process
        queue_entries = None
        if self.from_queue:
            try:
                filters = ProcessQueue.dataset_id == "SOLO_HK_PLATFORM"
                queue_entries = query_db(
                    self.session,
                    ProcessQueue,
                    filters=filters,
                    order_by=asc(ProcessQueue.start_time),
                    limit=self.limit,
                    tryouts=self.tryouts,
                    wait=self.wait,
                    to_dict="records",
                    raise_exception=True,
                )
            except Exception as e:
                logger.error("[DbToSoloHk]: Flagged current job as failed")
                logger.debug(e)
                # TODO - Add a method to make failed/ directory with relevant
                # files inside (log, input/output, etc.)
                return
            else:
                logger.debug(
                    f"{len(queue_entries)} dates to process returned from the database queue"
                )
                self.date_list = [
                    current_row["start_time"].date() for current_row in queue_entries
                ]
        elif not self.date_list:
            current_day = self.start_time
            while current_day < self.end_time:
                self.date_list.append(current_day.date())
                current_day += timedelta(days=1)

        # Make sure to have unique values in date list
        self.date_list = sorted(list(set(self.date_list)))

        n_dates = len(self.date_list)
        if n_dates == 0:
            logger.info("No date to process, exiting")
            return
        else:
            logger.info(f"Exporting SOLO HK from database for {n_dates} dates")

        # Loop on date_list
        output_files = []
        queue_entries_to_clean = []
        for i, current_date in enumerate(self.date_list):
            # Get file version
            if queue_entries and queue_entries[i]["version"]:
                current_version = valid_data_version(queue_entries[i]["version"])
            else:
                current_version = self.data_version

            # Generate the output XML file path
            xml_path = os.path.join(
                self.output_dir,
                f"solo_HK_platform_{current_date.strftime(TIME_DAILY_STRFORMAT)}_V{current_version}.xml",
            )
            # Open a file with the name
            if os.path.isfile(xml_path) and not self.overwrite:
                logger.info(f"{xml_path} already exists!")
                if queue_entries:
                    queue_entries_to_clean.append(queue_entries[i])
                continue

            # Current start is date at 00:00:00
            current_start = datetime.combine(current_date, datetime.min.time())
            # current end is date + 1 at 00:00:00
            current_end = datetime.combine(
                current_date + timedelta(days=1), datetime.min.time()
            )
            # Add input filters
            time = SoloHkParam.utc_time
            # greater or equal than ...
            # lesser than ...
            filters = (time >= str(current_start)) & (time < str(current_end))

            # Query data from database
            try:
                solohk_entries = query_db(
                    self.session,
                    SoloHkParam,
                    filters=filters,
                    limit=self.limit,
                    tryouts=self.tryouts,
                    wait=self.wait,
                    order_by=SoloHkParam.utc_time,
                    to_dict="records",
                    raise_exception=True,
                )
            except Exception as e:
                logger.error(f"[DbToSoloHk]: Task has failed for {current_date}")
                logger.debug(e)
                # TODO - Add a method to make failed/ directory with relevant
                # files inside (log, input/output, etc.)
                self.failed_files.append(xml_path)
                continue
            else:
                n_entries = len(solohk_entries)
                logger.info(
                    f"{n_entries} entries returned from the database for {current_date}"
                )

            if n_entries == 0:
                return

            # Write the output file using Jinja2
            # Load GSE test log template
            template = JENV.get_template(SOLO_HK_TEMPLATE)

            # Convert solo_hk_param table data into list of dictionaries
            param_sample_list = [
                self._convert_entry(current_entry) for current_entry in solohk_entries
            ]

            # Build the output SOLO HK XML template render
            render = template.render(param_sample_list=param_sample_list)

            logger.info(f"Saving {xml_path}...  ({n_dates - i - 1} remaining)")
            # Create output file
            with open(xml_path, "w") as outfile:
                outfile.write(render)
            if os.path.isfile(xml_path):
                output_files.append(xml_path)
            else:
                self.failed_files.append(xml_path)
                logger.error(f"Saving {xml_path} has failed")
                continue

            if queue_entries:
                queue_entries_to_clean.append(queue_entries[i])

        # Remove corresponding entry in pipeline.data_queue
        if self.clean_queue and queue_entries_to_clean:
            for current_entry in queue_entries_to_clean:
                if not delete_in_db(
                    self.session,
                    ProcessQueue,
                    filters=(ProcessQueue.id == current_entry["id"]),
                    tryouts=self.tryouts,
                    wait=self.wait,
                ):
                    logger.error(
                        f"{current_entry} has not been deleted from pipeline.data_queue table!"
                    )
                else:
                    logger.info(
                        f"{current_entry} deleted from pipeline.data_queue table"
                    )

        # store list of output files as output target of the task
        self.outputs["solo_hk_files"].filepath = output_files

    def _convert_entry(self, current_entry):
        """

        :param current_entry:
        :return:
        """
        xml_element = {}
        xml_element["Name"] = current_entry["name"]
        xml_element["TimeStampAsciiA"] = current_entry["utc_time"].strftime(
            TIME_SQL_STRFORMAT
        )
        xml_element["Unit"] = current_entry["unit"]
        xml_element["Description"] = current_entry["description"]
        xml_element["EngineeringValue"] = current_entry["eng_value"]
        xml_element["RawValue"] = current_entry["raw_value"]

        return xml_element
